"""
pint.facets.plain.registry
~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2022 by Pint Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.

The registry contains the following important methods:

- parse_unit_name: Parse a unit to identify prefix, unit name and suffix
  by walking the list of prefix and suffix.
  Result is cached: NO
- parse_units: Parse a units expression and returns a UnitContainer with
  the canonical names.
  The expression can only contain products, ratios and powers of units;
  prefixed units and pluralized units.
  Result is cached: YES
- parse_expression: Parse a mathematical expression including units and
  return a quantity object.
  Result is cached: NO

"""

from __future__ import annotations

import copy
import functools
import inspect
import itertools
import pathlib
import re
from collections import defaultdict
from collections.abc import Callable, Generator, Iterable, Iterator
from decimal import Decimal
from fractions import Fraction
from token import NAME, NUMBER
from tokenize import TokenInfo
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from ...compat import Locale
    from ..context import Context

    # from ..._typing import Quantity, Unit

import platformdirs

from ... import pint_eval
from ..._typing import (
    Handler,
    QuantityArgument,
    QuantityOrUnitLike,
    Scalar,
    UnitLike,
)
from ...compat import Self, TypeAlias, deprecated
from ...errors import (
    DimensionalityError,
    OffsetUnitCalculusError,
    RedefinitionError,
    UndefinedUnitError,
)
from ...pint_eval import build_eval_tree
from ...util import (
    ParserHelper,
    _is_dim,
    create_class_with_registry,
    getattr_maybe_raise,
    logger,
    solve_dependencies,
    string_preprocessor,
    to_units_container,
)
from ...util import UnitsContainer as UnitsContainer
from .definitions import (
    AliasDefinition,
    CommentDefinition,
    DefaultsDefinition,
    DerivedDimensionDefinition,
    DimensionDefinition,
    NamedDefinition,
    PrefixDefinition,
    UnitDefinition,
)
from .objects import PlainQuantity, PlainUnit

T = TypeVar("T")

_BLOCK_RE = re.compile(r"[ (]")


@functools.lru_cache
def pattern_to_regex(pattern: str | re.Pattern[str]) -> re.Pattern[str]:
    # TODO: This has been changed during typing improvements.
    # if hasattr(pattern, "finditer"):
    if not isinstance(pattern, str):
        pattern = pattern.pattern

    # Replace "{unit_name}" match string with float regex with unit_name as group
    pattern = re.sub(
        r"{(\w+)}", r"(?P<\1>[+-]?[0-9]+(?:.[0-9]+)?(?:[Ee][+-]?[0-9]+)?)", pattern
    )

    return re.compile(pattern)


NON_INT_TYPE = type[Union[float, Decimal, Fraction]]
PreprocessorType = Callable[[str], str]


class RegistryCache:
    """Cache to speed up unit registries"""

    def __init__(self) -> None:
        #: Maps dimensionality (UnitsContainer) to Units (str)
        self.dimensional_equivalents: dict[UnitsContainer, frozenset[str]] = {}

        #: Maps dimensionality (UnitsContainer) to Dimensionality (UnitsContainer)
        # TODO: this description is not right.
        self.root_units: dict[UnitsContainer, tuple[Scalar, UnitsContainer]] = {}

        #: Maps dimensionality (UnitsContainer) to Units (UnitsContainer)
        self.dimensionality: dict[UnitsContainer, UnitsContainer] = {}

        #: Cache the unit name associated to user input. ('mV' -> 'millivolt')
        self.parse_unit: dict[str, UnitsContainer] = {}

        self.conversion_factor: dict[
            tuple[UnitsContainer, UnitsContainer], Scalar | DimensionalityError
        ] = {}

    def __eq__(self, other: Any):
        if not isinstance(other, self.__class__):
            return False
        attrs = (
            "dimensional_equivalents",
            "root_units",
            "dimensionality",
            "parse_unit",
            "conversion_factor",
        )
        return all(getattr(self, attr) == getattr(other, attr) for attr in attrs)


class RegistryMeta(type):
    """This is just to call after_init at the right time
    instead of asking the developer to do it when subclassing.
    """

    def __call__(self, *args: Any, **kwargs: Any):
        obj = super().__call__(*args, **kwargs)
        obj._after_init()
        return obj


# Generic types used to mark types associated to Registries.
QuantityT = TypeVar("QuantityT", bound=PlainQuantity[Any])
UnitT = TypeVar("UnitT", bound=PlainUnit)


class GenericPlainRegistry(Generic[QuantityT, UnitT], metaclass=RegistryMeta):
    """Base class for all registries.

    Capabilities:

    - Register units, prefixes, and dimensions, and their relations.
    - Convert between units.
    - Find dimensionality of a unit.
    - Parse units with prefix and/or suffix.
    - Parse expressions.
    - Parse a definition file.
    - Allow extending the definition file parser by registering @ directives.

    Parameters
    ----------
    filename : str or None
        path of the units definition file to load or line iterable object. Empty to load
        the default definition file. None to leave the UnitRegistry empty.
    force_ndarray : bool
        convert any input, scalar or not to a numpy.ndarray.
    force_ndarray_like : bool
        convert all inputs other than duck arrays to a numpy.ndarray.
    on_redefinition : str
        action to take in case a unit is redefined: 'warn', 'raise', 'ignore'
    auto_reduce_dimensions :
        If True, reduce dimensionality on appropriate operations.
    autoconvert_to_preferred :
        If True, converts preferred units on appropriate operations.
    preprocessors :
        list of callables which are iteratively ran on any input expression or unit
        string
    fmt_locale :
        locale identifier string, used in `format_babel`
    non_int_type : type
        numerical type used for non integer values. (Default: float)
    case_sensitive : bool, optional
        Control default case sensitivity of unit parsing. (Default: True)
    cache_folder : str or pathlib.Path or None, optional
        Specify the folder in which cache files are saved and loaded from.
        If None, the cache is disabled. (default)
    separate_format_defaults : bool, optional
        Separate the default format into magnitude and unit formats as soon as
        possible. The deprecated default is not to separate. This will change in a
        future release.
    """

    Quantity: type[QuantityT]
    Unit: type[UnitT]

    _diskcache = None
    _def_parser = None

    def __init__(
        self,
        filename="",
        force_ndarray: bool = False,
        force_ndarray_like: bool = False,
        on_redefinition: str = "warn",
        auto_reduce_dimensions: bool = False,
        autoconvert_to_preferred: bool = False,
        preprocessors: list[PreprocessorType] | None = None,
        fmt_locale: str | None = None,
        non_int_type: NON_INT_TYPE = float,
        case_sensitive: bool = True,
        cache_folder: str | pathlib.Path | None = None,
        separate_format_defaults: bool | None = None,
        mpl_formatter: str = "{:P}",
    ):
        #: Map a definition class to a adder methods.
        self._adders: Handler = {}
        self._register_definition_adders()
        self._init_dynamic_classes()

        if cache_folder == ":auto:":
            cache_folder = platformdirs.user_cache_path(appname="pint", appauthor=False)

        from ... import delegates  # TODO: change this

        if cache_folder is not None:
            self._diskcache = delegates.build_disk_cache_class(non_int_type)(
                cache_folder
            )

        self._def_parser = delegates.txt_defparser.DefParser(
            delegates.ParserConfig(non_int_type), diskcache=self._diskcache
        )

        self.formatter = delegates.Formatter(self)
        self._filename = filename
        self.force_ndarray = force_ndarray
        self.force_ndarray_like = force_ndarray_like
        self.preprocessors = preprocessors or []
        # use a default preprocessor to support "%"
        self.preprocessors.insert(0, lambda string: string.replace("%", " percent "))

        # use a default preprocessor to support permille "‰"
        self.preprocessors.insert(0, lambda string: string.replace("‰", " permille "))

        # use a default preprocessor to support multiplication sign "×"
        self.preprocessors.insert(0, lambda string: string.replace("×", "*"))

        #: mode used to fill in the format defaults
        self.separate_format_defaults = separate_format_defaults

        #: Action to take in case a unit is redefined. 'warn', 'raise', 'ignore'
        self._on_redefinition = on_redefinition

        #: Determines if dimensionality should be reduced on appropriate operations.
        self.auto_reduce_dimensions = auto_reduce_dimensions

        #: Determines if units will be converted to preffered on appropriate operations.
        self.autoconvert_to_preferred = autoconvert_to_preferred

        #: Default locale identifier string, used when calling format_babel without explicit locale.
        self.formatter.set_locale(fmt_locale)

        #: sets the formatter used when plotting with matplotlib
        self.mpl_formatter = mpl_formatter

        #: Numerical type used for non integer values.
        self._non_int_type = non_int_type

        #: Default unit case sensitivity
        self.case_sensitive = case_sensitive

        #: Map between name (string) and value (string) of defaults stored in the
        #: definitions file.
        self._defaults: dict[str, str] = {}

        #: Map dimension name (string) to its definition (DimensionDefinition).
        self._dimensions: dict[
            str, DimensionDefinition | DerivedDimensionDefinition
        ] = {}

        #: Map unit name (string) to its definition (UnitDefinition).
        #: Might contain prefixed units.
        self._units: dict[str, UnitDefinition] = {}

        #: List base unit names
        self._base_units: list[str] = []

        #: Map unit name in lower case (string) to a set of unit names with the right
        #: case.
        #: Does not contain prefixed units.
        #: e.g: 'hz' - > set('Hz', )
        self._units_casei: dict[str, set[str]] = defaultdict(set)

        #: Map prefix name (string) to its definition (PrefixDefinition).
        self._prefixes: dict[str, PrefixDefinition] = {"": PrefixDefinition("", 1)}

        #: Map suffix name (string) to canonical , and unit alias to canonical unit name
        self._suffixes: dict[str, str] = {"": "", "s": ""}

        #: Map contexts to RegistryCache
        self._cache = RegistryCache()

        self._initialized = False

    def _init_dynamic_classes(self) -> None:
        """Generate subclasses on the fly and attach them to self"""

        self.Unit = create_class_with_registry(self, self.Unit)
        self.Quantity = create_class_with_registry(self, self.Quantity)

    def _after_init(self) -> None:
        """This should be called after all __init__"""

        if self._filename == "":
            path = pathlib.Path(__file__).parent.parent.parent / "default_en.txt"
            loaded_files = self.load_definitions(path, True)
        elif self._filename is not None:
            loaded_files = self.load_definitions(self._filename)
        else:
            loaded_files = None

        self._build_cache(loaded_files)
        self._initialized = True

    def _register_adder(
        self,
        definition_class: type[T],
        adder_func: Callable[
            [
                T,
            ],
            None,
        ],
    ) -> None:
        """Register a block definition."""
        self._adders[definition_class] = adder_func

    def _register_definition_adders(self) -> None:
        self._register_adder(AliasDefinition, self._add_alias)
        self._register_adder(DefaultsDefinition, self._add_defaults)
        self._register_adder(CommentDefinition, lambda o: o)
        self._register_adder(PrefixDefinition, self._add_prefix)
        self._register_adder(UnitDefinition, self._add_unit)
        self._register_adder(DimensionDefinition, self._add_dimension)
        self._register_adder(DerivedDimensionDefinition, self._add_derived_dimension)

    def __deepcopy__(self: Self, memo) -> type[Self]:
        new = object.__new__(type(self))
        new.__dict__ = copy.deepcopy(self.__dict__, memo)
        new._init_dynamic_classes()
        return new

    def __getattr__(self, item: str) -> UnitT:
        getattr_maybe_raise(self, item)

        # self.Unit will call parse_units
        return self.Unit(item)

    @deprecated(
        "Calling the getitem method from a UnitRegistry will be removed in future versions of pint.\n"
        "use `parse_expression` method or use the registry as a callable."
    )
    def __getitem__(self, item: str) -> UnitT:
        return self.parse_expression(item)

    def __contains__(self, item: str) -> bool:
        """Support checking prefixed units with the `in` operator"""
        try:
            self.__getattr__(item)
            return True
        except UndefinedUnitError:
            return False

    def __dir__(self) -> list[str]:
        #: Calling dir(registry) gives all units, methods, and attributes.
        #: Also used for autocompletion in IPython.
        return list(self._units.keys()) + list(object.__dir__(self))

    def __iter__(self) -> Iterator[str]:
        """Allows for listing all units in registry with `list(ureg)`.

        Returns
        -------
        Iterator over names of all units in registry, ordered alphabetically.
        """
        return iter(sorted(self._units.keys()))

    @property
    @deprecated(
        "This function will be removed in future versions of pint.\n"
        "Use ureg.formatter.fmt_locale"
    )
    def fmt_locale(self) -> Locale | None:
        return self.formatter.locale

    @fmt_locale.setter
    @deprecated(
        "This function will be removed in future versions of pint.\n"
        "Use ureg.formatter.set_locale"
    )
    def fmt_locale(self, loc: str | None):
        self.formatter.set_locale(loc)

    @deprecated(
        "This function will be removed in future versions of pint.\n"
        "Use ureg.formatter.set_locale"
    )
    def set_fmt_locale(self, loc: str | None) -> None:
        """Change the locale used by default by `format_babel`.

        Parameters
        ----------
        loc : str or None
            None` (do not translate), 'sys' (detect the system locale) or a locale id string.
        """

        self.formatter.set_locale(loc)

    @property
    @deprecated(
        "This function will be removed in future versions of pint.\n"
        "Use ureg.formatter.default_format"
    )
    def default_format(self) -> str:
        """Default formatting string for quantities."""
        return self.formatter.default_format

    @default_format.setter
    @deprecated(
        "This function will be removed in future versions of pint.\n"
        "Use ureg.formatter.default_format"
    )
    def default_format(self, value: str) -> None:
        self.formatter.default_format = value

    @property
    def cache_folder(self) -> pathlib.Path | None:
        if self._diskcache:
            return self._diskcache.cache_folder
        return None

    @property
    def non_int_type(self):
        return self._non_int_type

    def define(self, definition: str | type) -> None:
        """Add unit to the registry.

        Parameters
        ----------
        definition : str or Definition
            a dimension, unit or prefix definition.
        """

        if isinstance(definition, str):
            parsed_project = self._def_parser.parse_string(definition)

            for definition in self._def_parser.iter_parsed_project(parsed_project):
                self._helper_dispatch_adder(definition)
        else:
            self._helper_dispatch_adder(definition)

    ############
    # Adders
    # - we first provide some helpers that deal with repetitive task.
    # - then we define specific adder for each definition class. :-D
    ############

    def _helper_dispatch_adder(self, definition: Any) -> None:
        """Helper function to add a single definition,
        choosing the appropiate method by class.
        """
        for cls in inspect.getmro(definition.__class__):
            if cls in self._adders:
                adder_func = self._adders[cls]
                break
        else:
            raise TypeError(
                f"No loader function defined for {definition.__class__.__name__}"
            )

        adder_func(definition)

    def _helper_adder(
        self,
        definition: NamedDefinition,
        target_dict: dict[str, Any],
        casei_target_dict: dict[str, Any] | None,
    ) -> None:
        """Helper function to store a definition in the internal dictionaries.
        It stores the definition under its name, symbol and aliases.
        """
        self._helper_single_adder(
            definition.name, definition, target_dict, casei_target_dict
        )

        # TODO: Not sure why but using hasattr does not work here.
        if getattr(definition, "has_symbol", ""):
            self._helper_single_adder(
                definition.symbol, definition, target_dict, casei_target_dict
            )

        for alias in getattr(definition, "aliases", ()):
            if " " in alias:
                logger.warn("Alias cannot contain a space: " + alias)

            self._helper_single_adder(alias, definition, target_dict, casei_target_dict)

    def _helper_single_adder(
        self,
        key: str,
        value: NamedDefinition,
        target_dict: dict[str, Any],
        casei_target_dict: dict[str, Any] | None,
    ) -> None:
        """Helper function to store a definition in the internal dictionaries.

        It warns or raise error on redefinition.
        """
        if key in target_dict:
            if self._on_redefinition == "raise":
                raise RedefinitionError(key, type(value))
            elif self._on_redefinition == "warn":
                logger.warning(f"Redefining '{key}' ({type(value)})")

        target_dict[key] = value
        if casei_target_dict is not None:
            casei_target_dict[key.lower()].add(key)

    def _add_defaults(self, defaults_definition: DefaultsDefinition) -> None:
        for k, v in defaults_definition.items():
            self._defaults[k] = v

    def _add_alias(self, definition: AliasDefinition) -> None:
        unit_dict = self._units
        unit = unit_dict[definition.name]
        while not isinstance(unit, UnitDefinition):
            unit = unit_dict[unit.name]
        for alias in definition.aliases:
            self._helper_single_adder(alias, unit, self._units, self._units_casei)

    def _add_dimension(self, definition: DimensionDefinition) -> None:
        self._helper_adder(definition, self._dimensions, None)

    def _add_derived_dimension(self, definition: DerivedDimensionDefinition) -> None:
        for dim_name in definition.reference.keys():
            if dim_name not in self._dimensions:
                self._add_dimension(DimensionDefinition(dim_name))
        self._helper_adder(definition, self._dimensions, None)

    def _add_prefix(self, definition: PrefixDefinition) -> None:
        self._helper_adder(definition, self._prefixes, None)

    def _add_unit(self, definition: UnitDefinition) -> None:
        if definition.is_base:
            self._base_units.append(definition.name)
            for dim_name in definition.reference.keys():
                if dim_name not in self._dimensions:
                    self._add_dimension(DimensionDefinition(dim_name))

        self._helper_adder(definition, self._units, self._units_casei)

    def load_definitions(
        self, file: Iterable[str] | str | pathlib.Path, is_resource: bool = False
    ):
        """Add units and prefixes defined in a definition text file.

        Parameters
        ----------
        file :
            can be a filename or a line iterable.
        is_resource :
            used to indicate that the file is a resource file
            and therefore should be loaded from the package. (Default value = False)
        """

        if isinstance(file, (list, tuple)):
            # TODO: this hack was to keep it backwards compatible.
            parsed_project = self._def_parser.parse_string("\n".join(file))
        else:
            parsed_project = self._def_parser.parse_file(file)

        for definition in self._def_parser.iter_parsed_project(parsed_project):
            self._helper_dispatch_adder(definition)

        return parsed_project

    def _build_cache(self, loaded_files=None) -> None:
        """Build a cache of dimensionality and plain units."""

        diskcache = self._diskcache
        if loaded_files and diskcache:
            cache, cache_basename = diskcache.load(loaded_files, "build_cache")
            if cache is None:
                self._build_cache()
                diskcache.save(self._cache, loaded_files, "build_cache")
            return

        self._cache = RegistryCache()

        deps: dict[str, set[str]] = {
            name: set(definition.reference.keys()) if definition.reference else set()
            for name, definition in self._units.items()
        }

        for unit_names in solve_dependencies(deps):
            for unit_name in unit_names:
                if "[" in unit_name:
                    continue
                parsed_names = self.parse_unit_name(unit_name)
                if parsed_names:
                    prefix, base_name, _ = parsed_names[0]
                else:
                    prefix, base_name = "", unit_name

                try:
                    uc = ParserHelper.from_word(base_name, self.non_int_type)

                    bu = self._get_root_units(uc)
                    di = self._get_dimensionality(uc)

                    self._cache.root_units[uc] = bu
                    self._cache.dimensionality[uc] = di

                    if not prefix:
                        dimeq_set = self._cache.dimensional_equivalents.setdefault(
                            di, set()
                        )
                        dimeq_set.add(self._units[base_name].name)

                except Exception as exc:
                    logger.warning(f"Could not resolve {unit_name}: {exc!r}")
        return self._cache

    def get_name(self, name_or_alias: str, case_sensitive: bool | None = None) -> str:
        """Return the canonical name of a unit."""

        if name_or_alias == "dimensionless":
            return ""

        try:
            return self._units[name_or_alias].name
        except KeyError:
            pass

        candidates = self.parse_unit_name(name_or_alias, case_sensitive)
        if not candidates:
            raise UndefinedUnitError(name_or_alias)

        prefix, unit_name, _ = candidates[0]
        if len(candidates) > 1:
            logger.warning(
                f"Parsing {name_or_alias} yield multiple results. Options are: {candidates!r}"
            )

        if prefix:
            if not self._units[unit_name].is_multiplicative:
                raise OffsetUnitCalculusError(
                    "Prefixing a unit requires multiplying the unit."
                )

            name = prefix + unit_name
            symbol = self.get_symbol(name, case_sensitive)
            prefix_def = self._prefixes[prefix]
            self._units[name] = UnitDefinition(
                name,
                symbol,
                tuple(),
                prefix_def.converter,
                self.UnitsContainer({unit_name: 1}),
            )
            return prefix + unit_name

        return unit_name

    def get_symbol(self, name_or_alias: str, case_sensitive: bool | None = None) -> str:
        """Return the preferred alias for a unit."""
        candidates = self.parse_unit_name(name_or_alias, case_sensitive)
        if not candidates:
            raise UndefinedUnitError(name_or_alias)

        prefix, unit_name, _ = candidates[0]
        if len(candidates) > 1:
            logger.warning(
                f"Parsing {name_or_alias} yield multiple results. Options are: {candidates!r}"
            )

        return self._prefixes[prefix].symbol + self._units[unit_name].symbol

    def _get_symbol(self, name: str) -> str:
        return self._units[name].symbol

    def get_dimensionality(self, input_units: UnitLike) -> UnitsContainer:
        """Convert unit or dict of units or dimensions to a dict of plain dimensions
        dimensions
        """

        # TODO: This should be to_units_container(input_units, self)
        # but this tries to reparse and fail for dimensions.
        input_units = to_units_container(input_units)

        return self._get_dimensionality(input_units)

    def _get_dimensionality(self, input_units: UnitsContainer | None) -> UnitsContainer:
        """Convert a UnitsContainer to plain dimensions."""
        if not input_units:
            return self.UnitsContainer()

        cache = self._cache.dimensionality

        try:
            return cache[input_units]
        except KeyError:
            pass

        accumulator: dict[str, int] = defaultdict(int)
        self._get_dimensionality_recurse(input_units, 1, accumulator)

        if "[]" in accumulator:
            del accumulator["[]"]

        dims = self.UnitsContainer({k: v for k, v in accumulator.items() if v != 0})

        cache[input_units] = dims

        return dims

    def _get_dimensionality_recurse(
        self, ref: UnitsContainer, exp: Scalar, accumulator: dict[str, int]
    ) -> None:
        for key in ref:
            exp2 = exp * ref[key]
            if _is_dim(key):
                try:
                    reg = self._dimensions[key]
                except KeyError:
                    raise ValueError(
                        f"{key} is not defined as dimension in the pint UnitRegistry"
                    )
                if isinstance(reg, DerivedDimensionDefinition):
                    self._get_dimensionality_recurse(reg.reference, exp2, accumulator)
                else:
                    # DimensionDefinition.
                    accumulator[key] += exp2

            else:
                reg = self._units[self.get_name(key)]
                if reg.reference is not None:
                    self._get_dimensionality_recurse(reg.reference, exp2, accumulator)

    def _get_dimensionality_ratio(
        self, unit1: UnitLike, unit2: UnitLike
    ) -> Scalar | None:
        """Get the exponential ratio between two units, i.e. solve unit2 = unit1**x for x.

        Parameters
        ----------
        unit1 : UnitsContainer compatible (str, Unit, UnitsContainer, dict)
            first unit
        unit2 : UnitsContainer compatible (str, Unit, UnitsContainer, dict)
            second unit

        Returns
        -------
        number or None
            exponential proportionality or None if the units cannot be converted

        """
        # shortcut in case of equal units
        if unit1 == unit2:
            return 1

        dim1, dim2 = (self.get_dimensionality(unit) for unit in (unit1, unit2))
        if dim1 == dim2:
            return 1
        elif not dim1 or not dim2 or dim1.keys() != dim2.keys():  # not comparable
            return None

        ratios = (dim2[key] / val for key, val in dim1.items())
        first = next(ratios)
        if all(r == first for r in ratios):  # all are same, we're good
            return first
        return None

    def get_root_units(
        self, input_units: UnitLike, check_nonmult: bool = True
    ) -> tuple[Scalar, UnitT]:
        """Convert unit or dict of units to the root units.

        If any unit is non multiplicative and check_converter is True,
        then None is returned as the multiplicative factor.

        Parameters
        ----------
        input_units : UnitsContainer or str
            units
        check_nonmult : bool
            if True, None will be returned as the
            multiplicative factor if a non-multiplicative
            units is found in the final Units. (Default value = True)

        Returns
        -------
        Number, pint.Unit
            multiplicative factor, plain units

        """
        input_units = to_units_container(input_units, self)

        f, units = self._get_root_units(input_units, check_nonmult)

        return f, self.Unit(units)

    def _get_conversion_factor(
        self, src: UnitsContainer, dst: UnitsContainer
    ) -> Scalar | DimensionalityError:
        """Get conversion factor in non-multiplicative units.

        Parameters
        ----------
        src
            Source units
        dst
            Target units

        Returns
        -------
            Conversion factor or DimensionalityError
        """
        cache = self._cache.conversion_factor
        try:
            return cache[(src, dst)]
        except KeyError:
            pass

        src_dim = self._get_dimensionality(src)
        dst_dim = self._get_dimensionality(dst)

        # If the source and destination dimensionality are different,
        # then the conversion cannot be performed.
        if src_dim != dst_dim:
            return DimensionalityError(src, dst, src_dim, dst_dim)

        # Here src and dst have only multiplicative units left. Thus we can
        # convert with a factor.
        factor, _ = self._get_root_units(src / dst)

        cache[(src, dst)] = factor
        return factor

    def _get_root_units(
        self, input_units: UnitsContainer, check_nonmult: bool = True
    ) -> tuple[Scalar, UnitsContainer]:
        """Convert unit or dict of units to the root units.

        If any unit is non multiplicative and check_converter is True,
        then None is returned as the multiplicative factor.

        Parameters
        ----------
        input_units : UnitsContainer or dict
            units
        check_nonmult : bool
            if True, None will be returned as the
            multiplicative factor if a non-multiplicative
            units is found in the final Units. (Default value = True)

        Returns
        -------
        number, Unit
            multiplicative factor, plain units

        """
        if not input_units:
            return 1, self.UnitsContainer()

        cache = self._cache.root_units
        try:
            return cache[input_units]
        except KeyError:
            pass

        accumulators: dict[str | None, int] = defaultdict(int)
        fraction: dict[str, dict[Decimal | float, Decimal | int]] = dict(
            numerator=dict(), denominator=dict()
        )
        self._get_root_units_recurse(input_units, 1, accumulators, fraction)

        # Identify if terms appear in both numerator and denominator
        def terms_are_unique(fraction):
            for n_factor, n_exp in fraction["numerator"].items():
                if n_factor in fraction["denominator"]:
                    return False
            return True

        # Cancel out terms where factor matches
        while not terms_are_unique(fraction):
            for n_factor, n_exponent in fraction["numerator"].items():
                if n_factor in fraction["denominator"]:
                    if n_exponent >= fraction["denominator"][n_factor]:
                        fraction["numerator"][n_factor] -= fraction["denominator"][
                            n_factor
                        ]
                        del fraction["denominator"][n_factor]
                        continue
            for d_factor, d_exponent in fraction["denominator"].items():
                if d_factor in fraction["numerator"]:
                    if d_exponent >= fraction["numerator"][d_factor]:
                        fraction["denominator"][d_factor] -= fraction["numerator"][
                            d_factor
                        ]
                        del fraction["numerator"][d_factor]
                        continue

        factor = 1
        for n_factor, n_exponent in fraction["numerator"].copy().items():
            if n_exponent == 0:
                del fraction["numerator"][n_factor]
            else:
                factor *= n_factor**n_exponent
        for d_factor, d_exponent in fraction["denominator"].copy().items():
            if d_exponent == 0:
                del fraction["denominator"][d_factor]
            else:
                factor *= d_factor**-d_exponent

        units = self.UnitsContainer(
            {k: v for k, v in accumulators.items() if k is not None and v != 0}
        )

        # Check if any of the final units is non multiplicative and return None instead.
        if check_nonmult:
            if any(not self._units[unit].converter.is_multiplicative for unit in units):
                factor = None

        cache[input_units] = factor, units
        return factor, units

    def get_base_units(
        self,
        input_units: UnitsContainer | str,
        check_nonmult: bool = True,
        system=None,
    ) -> tuple[Scalar, UnitT]:
        """Convert unit or dict of units to the plain units.

        If any unit is non multiplicative and check_converter is True,
        then None is returned as the multiplicative factor.

        Parameters
        ----------
        input_units : UnitsContainer or str
            units
        check_nonmult : bool
            If True, None will be returned as the multiplicative factor if
            non-multiplicative units are found in the final Units.
            (Default value = True)
        system :
             (Default value = None)

        Returns
        -------
        Number, pint.Unit
            multiplicative factor, plain units

        """

        return self.get_root_units(input_units, check_nonmult)

    # TODO: accumulators breaks typing list[int, dict[str, int]]
    # So we have changed the behavior here
    def _get_root_units_recurse(
        self,
        ref: UnitsContainer,
        exp: Scalar,
        accumulators: dict[str | None, int],
        fraction: dict[str, dict[Decimal | float, Decimal | int]],
    ) -> None:
        """

        accumulators None keeps the scalar prefactor not associated with a specific unit.

        """
        for key in ref:
            exp2 = exp * ref[key]
            key = self.get_name(key)
            reg = self._units[key]
            if reg.is_base:
                accumulators[key] += exp2
            else:
                # Build numerator and denominator
                if exp2 < 0:
                    fraction["denominator"][reg.converter.scale] = (
                        fraction["denominator"].get(reg.converter.scale, 0) - exp2
                    )
                else:
                    fraction["numerator"][reg.converter.scale] = (
                        fraction["numerator"].get(reg.converter.scale, 0) + exp2
                    )
                if reg.reference is not None:
                    self._get_root_units_recurse(
                        reg.reference, exp2, accumulators, fraction
                    )

    def get_compatible_units(self, input_units: QuantityOrUnitLike) -> frozenset[UnitT]:
        """ """
        input_units = to_units_container(input_units)

        equiv = self._get_compatible_units(input_units)

        return frozenset(self.Unit(eq) for eq in equiv)

    def _get_compatible_units(
        self, input_units: UnitsContainer, *args, **kwargs
    ) -> frozenset[str]:
        """ """
        if not input_units:
            return frozenset()

        src_dim = self._get_dimensionality(input_units)
        return self._cache.dimensional_equivalents.setdefault(src_dim, frozenset())

    # TODO: remove context from here
    def is_compatible_with(
        self, obj1: Any, obj2: Any, *contexts: str | Context, **ctx_kwargs
    ) -> bool:
        """check if the other object is compatible

        Parameters
        ----------
        obj1, obj2
            The objects to check against each other. Treated as
            dimensionless if not a Quantity, Unit or str.
        *contexts : str or pint.Context
            Contexts to use in the transformation.
        **ctx_kwargs :
            Values for the Context/s

        Returns
        -------
        bool
        """
        if isinstance(obj1, (self.Quantity, self.Unit)):
            return obj1.is_compatible_with(obj2, *contexts, **ctx_kwargs)

        if isinstance(obj1, str):
            return self.parse_expression(obj1).is_compatible_with(
                obj2, *contexts, **ctx_kwargs
            )

        return not isinstance(obj2, (self.Quantity, self.Unit))

    def convert(
        self,
        value: T,
        src: QuantityOrUnitLike,
        dst: QuantityOrUnitLike,
        inplace: bool = False,
        **ctx_kwargs,
    ) -> T:
        """Convert value from some source to destination units.

        Parameters
        ----------
        value :
            value
        src : pint.Quantity or str
            source units.
        dst : pint.Quantity or str
            destination units.
        inplace :
             (Default value = False)

        Returns
        -------
        type
            converted value

        """
        src = to_units_container(src, self)

        dst = to_units_container(dst, self)

        if src == dst:
            return value

        return self._convert(value, src, dst, inplace, **ctx_kwargs)

    def _convert(
        self,
        value: T,
        src: UnitsContainer,
        dst: UnitsContainer,
        inplace: bool = False,
        check_dimensionality: bool = True,
    ) -> T:
        """Convert value from some source to destination units.

        Parameters
        ----------
        value :
            value
        src : UnitsContainer
            source units.
        dst : UnitsContainer
            destination units.
        inplace :
             (Default value = False)
        check_dimensionality :
             (Default value = True)

        Returns
        -------
        type
            converted value

        """

        factor = self._get_conversion_factor(src, dst)

        if isinstance(factor, DimensionalityError):
            raise factor

        # factor is type float and if our magnitude is type Decimal then
        # must first convert to Decimal before we can '*' the values
        if isinstance(value, Decimal):
            factor = Decimal(str(factor))
        elif isinstance(value, Fraction):
            factor = Fraction(str(factor))

        if inplace:
            value *= factor
        else:
            value = value * factor

        return value

    def parse_unit_name(
        self, unit_name: str, case_sensitive: bool | None = None
    ) -> tuple[tuple[str, str, str], ...]:
        """Parse a unit to identify prefix, unit name and suffix
        by walking the list of prefix and suffix.
        In case of equivalent combinations (e.g. ('kilo', 'gram', '') and
        ('', 'kilogram', ''), prefer those with prefix.

        Parameters
        ----------
        unit_name :

        case_sensitive : bool or None
            Control if unit lookup is case sensitive. Defaults to None, which uses the
            registry's case_sensitive setting

        Returns
        -------
        tuple of tuples (str, str, str)
            all non-equivalent combinations of (prefix, unit name, suffix)
        """

        case_sensitive = (
            self.case_sensitive if case_sensitive is None else case_sensitive
        )
        return self._dedup_candidates(
            self._yield_unit_triplets(unit_name, case_sensitive)
        )

    def _yield_unit_triplets(
        self, unit_name: str, case_sensitive: bool
    ) -> Generator[tuple[str, str, str]]:
        """Helper of parse_unit_name."""

        stw = unit_name.startswith
        edw = unit_name.endswith
        for suffix, prefix in itertools.product(self._suffixes, self._prefixes):
            if stw(prefix) and edw(suffix):
                name = unit_name[len(prefix) :]
                if suffix:
                    name = name[: -len(suffix)]
                    if len(name) == 1:
                        continue
                if case_sensitive:
                    if name in self._units:
                        yield (
                            self._prefixes[prefix].name,
                            self._units[name].name,
                            self._suffixes[suffix],
                        )
                else:
                    for real_name in self._units_casei.get(name.lower(), ()):
                        yield (
                            self._prefixes[prefix].name,
                            self._units[real_name].name,
                            self._suffixes[suffix],
                        )

    # TODO: keep this for backward compatibility
    _parse_unit_name = _yield_unit_triplets

    @staticmethod
    def _dedup_candidates(
        candidates: Iterable[tuple[str, str, str]],
    ) -> tuple[tuple[str, str, str], ...]:
        """Helper of parse_unit_name.

        Given an iterable of unit triplets (prefix, name, suffix), remove those with
        different names but equal value, preferring those with a prefix.

        e.g. ('kilo', 'gram', '') and ('', 'kilogram', '')
        """
        candidates = dict.fromkeys(candidates)  # ordered set
        for cp, cu, cs in list(candidates):
            assert isinstance(cp, str)
            assert isinstance(cu, str)
            if cs != "":
                raise NotImplementedError("non-empty suffix")
            if cp:
                candidates.pop(("", cp + cu, ""), None)
        return tuple(candidates)

    def parse_units(
        self,
        input_string: str,
        as_delta: bool | None = None,
        case_sensitive: bool | None = None,
    ) -> UnitT:
        """Parse a units expression and returns a UnitContainer with
        the canonical names.

        The expression can only contain products, ratios and powers of units.

        Parameters
        ----------
        input_string : str
        as_delta : bool or None
            if the expression has multiple units, the parser will
            interpret non multiplicative units as their `delta_` counterparts. (Default value = None)
        case_sensitive : bool or None
            Control if unit parsing is case sensitive. Defaults to None, which uses the
            registry's setting.

        Returns
        -------
            pint.Unit

        """

        return self.Unit(
            self.parse_units_as_container(input_string, as_delta, case_sensitive)
        )

    def parse_units_as_container(
        self,
        input_string: str,
        as_delta: bool | None = None,
        case_sensitive: bool | None = None,
    ) -> UnitsContainer:
        as_delta = (
            as_delta if as_delta is not None else True
        )  # TODO This only exists in nonmultiplicative
        case_sensitive = (
            case_sensitive if case_sensitive is not None else self.case_sensitive
        )
        return self._parse_units_as_container(input_string, as_delta, case_sensitive)

    def _parse_units_as_container(
        self,
        input_string: str,
        as_delta: bool = True,
        case_sensitive: bool = True,
    ) -> UnitsContainer:
        """Parse a units expression and returns a UnitContainer with
        the canonical names.
        """

        cache = self._cache.parse_unit
        # Issue #1097: it is possible, when a unit was defined while a different context
        # was active, that the unit is in self._cache.parse_unit but not in self._units.
        # If this is the case, force self._units to be repopulated.
        if as_delta and input_string in cache and input_string in self._units:
            return cache[input_string]

        for p in self.preprocessors:
            input_string = p(input_string)

        if not input_string:
            return self.UnitsContainer()

        # Sanitize input_string with whitespaces.
        input_string = input_string.strip()

        units = ParserHelper.from_string(input_string, self.non_int_type)
        if units.scale != 1:
            raise ValueError("Unit expression cannot have a scaling factor.")

        ret = self.UnitsContainer({})
        many = len(units) > 1
        for name in units:
            cname = self.get_name(name, case_sensitive=case_sensitive)
            value = units[name]
            if not cname:
                continue
            if as_delta and (many or (not many and value != 1)):
                definition = self._units[cname]
                if not definition.is_multiplicative:
                    cname = "delta_" + cname
            ret = ret.add(cname, value)

        if as_delta:
            cache[input_string] = ret

        return ret

    def _eval_token(
        self,
        token: TokenInfo,
        case_sensitive: bool | None = None,
        **values: QuantityArgument,
    ):
        """Evaluate a single token using the following rules:

        1. numerical values as strings are replaced by their numeric counterparts
            - integers are parsed as integers
            - other numeric values are parses of non_int_type
        2. strings in (inf, infinity, nan, dimensionless) with their numerical value.
        3. strings in values.keys() are replaced by Quantity(values[key])
        4. in other cases, the values are parsed as units and replaced by their canonical name.

        Parameters
        ----------
        token
            Token to evaluate.
        case_sensitive, optional
            If true, a case sensitive matching of the unit name will be done in the registry.
            If false, a case INsensitive matching of the unit name will be done in the registry.
            (Default value = None, which uses registry setting)
        **values
            Other string that will be parsed using the Quantity constructor on their corresponding value.
        """
        token_type = token[0]
        token_text = token[1]
        if token_type == NAME:
            if token_text == "dimensionless":
                return self.Quantity(1)
            elif token_text.lower() in ("inf", "infinity"):
                return self.non_int_type("inf")
            elif token_text.lower() == "nan":
                return self.non_int_type("nan")
            elif token_text in values:
                return self.Quantity(values[token_text])
            else:
                return self.Quantity(
                    1,
                    self.UnitsContainer(
                        {self.get_name(token_text, case_sensitive=case_sensitive): 1}
                    ),
                )
        elif token_type == NUMBER:
            return ParserHelper.eval_token(token, non_int_type=self.non_int_type)
        else:
            raise Exception("unknown token type")

    def parse_pattern(
        self,
        input_string: str,
        pattern: str,
        case_sensitive: bool | None = None,
        many: bool = False,
    ) -> list[str] | str | None:
        """Parse a string with a given regex pattern and returns result.

        Parameters
        ----------
        input_string

        pattern_string:
            The regex parse string
        case_sensitive, optional
            If true, a case sensitive matching of the unit name will be done in the registry.
            If false, a case INsensitive matching of the unit name will be done in the registry.
            (Default value = None, which uses registry setting)
        many, optional
             Match many results
             (Default value = False)
        """

        if not input_string:
            return [] if many else None

        # Parse string
        regex = pattern_to_regex(pattern)
        matched = re.finditer(regex, input_string)

        # Extract result(s)
        results = []
        for match in matched:
            # Extract units from result
            match = match.groupdict()

            # Parse units
            units = [
                float(value) * self.parse_expression(unit, case_sensitive)
                for unit, value in match.items()
            ]

            # Add to results
            results.append(units)

            # Return first match only
            if not many:
                return results[0]

        return results

    def parse_expression(
        self: Self,
        input_string: str,
        case_sensitive: bool | None = None,
        **values: QuantityArgument,
    ) -> QuantityT:
        """Parse a mathematical expression including units and return a quantity object.

        Numerical constants can be specified as keyword arguments and will take precedence
        over the names defined in the registry.

        Parameters
        ----------
        input_string

        case_sensitive, optional
            If true, a case sensitive matching of the unit name will be done in the registry.
            If false, a case INsensitive matching of the unit name will be done in the registry.
            (Default value = None, which uses registry setting)
        **values
            Other string that will be parsed using the Quantity constructor on their corresponding value.
        """
        if not input_string:
            return self.Quantity(1)

        for p in self.preprocessors:
            input_string = p(input_string)
        input_string = string_preprocessor(input_string)
        gen = pint_eval.tokenizer(input_string)

        def _define_op(s: str):
            return self._eval_token(s, case_sensitive=case_sensitive, **values)

        return build_eval_tree(gen).evaluate(_define_op)

    # We put this last to avoid overriding UnitsContainer
    # and I do not want to rename it.
    # TODO: Maybe in the future we need to change it to a more meaningful
    # non-colliding name.
    def UnitsContainer(self, *args: Any, **kwargs: Any) -> UnitsContainer:
        return UnitsContainer(*args, non_int_type=self.non_int_type, **kwargs)

    __call__ = parse_expression


class PlainRegistry(GenericPlainRegistry[PlainQuantity[Any], PlainUnit]):
    Quantity: TypeAlias = PlainQuantity[Any]
    Unit: TypeAlias = PlainUnit



"""
pint.delegates.formatter.html
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Implements:
- HTML: suitable for web/jupyter notebook outputs.

:copyright: 2022 by Pint Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import re
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from ..._typing import Magnitude
from ...compat import Unpack, ndarray, np
from ...util import iterable
from ._compound_unit_helpers import (
    BabelKwds,
    SortFunc,
    localize_per,
    prepare_compount_unit,
)
from ._format_helpers import (
    formatter,
    join_mu,
    join_unc,
    override_locale,
)
from ._spec_helpers import (
    remove_custom_flags,
    split_format,
)
from .plain import BaseFormatter

if TYPE_CHECKING:
    from ...facets.measurement import Measurement
    from ...facets.plain import MagnitudeT, PlainQuantity, PlainUnit

_EXP_PATTERN = re.compile(r"([0-9]\.?[0-9]*)e(-?)\+?0*([0-9]+)")


class HTMLFormatter(BaseFormatter):
    """HTML localizable text formatter."""

    def format_magnitude(
        self, magnitude: Magnitude, mspec: str = "", **babel_kwds: Unpack[BabelKwds]
    ) -> str:
        with override_locale(mspec, babel_kwds.get("locale", None)) as format_number:
            if hasattr(magnitude, "_repr_html_"):
                # If magnitude has an HTML repr, nest it within Pint's
                mstr = magnitude._repr_html_()  # type: ignore
                assert isinstance(mstr, str)
            else:
                if isinstance(magnitude, ndarray):
                    # Need to override for scalars, which are detected as iterable,
                    # and don't respond to printoptions.
                    if magnitude.ndim == 0:
                        mstr = format_number(magnitude)
                    else:
                        with np.printoptions(formatter={"float_kind": format_number}):
                            mstr = (
                                "<pre>" + format(magnitude).replace("\n", "") + "</pre>"
                            )
                elif not iterable(magnitude):
                    # Use plain text for scalars
                    mstr = format_number(magnitude)
                else:
                    # Use monospace font for other array-likes
                    mstr = (
                        "<pre>"
                        + format_number(magnitude).replace("\n", "<br>")
                        + "</pre>"
                    )

        m = _EXP_PATTERN.match(mstr)
        _exp_formatter = lambda s: f"<sup>{s}</sup>"

        if m:
            exp = int(m.group(2) + m.group(3))
            mstr = _EXP_PATTERN.sub(r"\1×10" + _exp_formatter(exp), mstr)

        return mstr

    def format_unit(
        self,
        unit: PlainUnit | Iterable[tuple[str, Any]],
        uspec: str = "",
        sort_func: SortFunc | None = None,
        **babel_kwds: Unpack[BabelKwds],
    ) -> str:
        numerator, denominator = prepare_compount_unit(
            unit,
            uspec,
            sort_func=sort_func,
            **babel_kwds,
            registry=self._registry,
        )

        if babel_kwds.get("locale", None):
            length = babel_kwds.get("length") or ("short" if "~" in uspec else "long")
            division_fmt = localize_per(length, babel_kwds.get("locale"), "{}/{}")
        else:
            division_fmt = "{}/{}"

        return formatter(
            numerator,
            denominator,
            as_ratio=True,
            single_denominator=True,
            product_fmt=r" ",
            division_fmt=division_fmt,
            power_fmt=r"{}<sup>{}</sup>",
            parentheses_fmt=r"({})",
        )

    def format_quantity(
        self,
        quantity: PlainQuantity[MagnitudeT],
        qspec: str = "",
        sort_func: SortFunc | None = None,
        **babel_kwds: Unpack[BabelKwds],
    ) -> str:
        registry = self._registry

        mspec, uspec = split_format(
            qspec, registry.formatter.default_format, registry.separate_format_defaults
        )

        if iterable(quantity.magnitude):
            # Use HTML table instead of plain text template for array-likes
            joint_fstring = (
                "<table><tbody>"
                "<tr><th>Magnitude</th>"
                "<td style='text-align:left;'>{}</td></tr>"
                "<tr><th>Units</th><td style='text-align:left;'>{}</td></tr>"
                "</tbody></table>"
            )
        else:
            joint_fstring = "{} {}"

        return join_mu(
            joint_fstring,
            self.format_magnitude(quantity.magnitude, mspec, **babel_kwds),
            self.format_unit(quantity.unit_items(), uspec, sort_func, **babel_kwds),
        )

    def format_uncertainty(
        self,
        uncertainty,
        unc_spec: str = "",
        sort_func: SortFunc | None = None,
        **babel_kwds: Unpack[BabelKwds],
    ) -> str:
        unc_str = format(uncertainty, unc_spec).replace("+/-", " &plusmn; ")

        unc_str = re.sub(r"\)e\+0?(\d+)", r")×10<sup>\1</sup>", unc_str)
        unc_str = re.sub(r"\)e-0?(\d+)", r")×10<sup>-\1</sup>", unc_str)
        return unc_str

    def format_measurement(
        self,
        measurement: Measurement,
        meas_spec: str = "",
        sort_func: SortFunc | None = None,
        **babel_kwds: Unpack[BabelKwds],
    ) -> str:
        registry = self._registry

        mspec, uspec = split_format(
            meas_spec,
            registry.formatter.default_format,
            registry.separate_format_defaults,
        )

        unc_spec = remove_custom_flags(meas_spec)

        joint_fstring = "{} {}"

        return join_unc(
            joint_fstring,
            "(",
            ")",
            self.format_uncertainty(measurement.magnitude, unc_spec, **babel_kwds),
            self.format_unit(measurement.units, uspec, sort_func, **babel_kwds),
        )



"""
pint.facets.system
~~~~~~~~~~~~~~~~~~

Adds pint the capability to system of units.

:copyright: 2022 by Pint Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from .definitions import SystemDefinition
from .objects import System
from .registry import GenericSystemRegistry, SystemRegistry

__all__ = ["SystemDefinition", "System", "SystemRegistry", "GenericSystemRegistry"]



"""
pint.delegates.formatter
~~~~~~~~~~~~~~~~~~~~~~~~

Easy to replace and extend string formatting.

See pint.delegates.formatter.plain.DefaultFormatter for a
description of a formatter.

:copyright: 2022 by Pint Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from .full import FullFormatter


class Formatter(FullFormatter):
    """Default Pint Formatter"""

    pass


__all__ = [
    "Formatter",
]



"""
pint.delegates.txt_defparser.system
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2022 by Pint Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import re
import typing as ty
from dataclasses import dataclass

import flexparser as fp

from ...facets.system import definitions
from ..base_defparser import PintParsedStatement
from . import block, common, plain


@dataclass(frozen=True)
class BaseUnitRule(PintParsedStatement, definitions.BaseUnitRule):
    @classmethod
    def from_string(cls, s: str) -> fp.NullableParsedResult[BaseUnitRule]:
        if ":" not in s:
            return cls(s.strip())
        parts = [p.strip() for p in s.split(":")]
        if len(parts) != 2:
            return common.DefinitionSyntaxError(
                f"Exactly two terms expected for rule, not {len(parts)} (`{s}`)"
            )
        return cls(*parts)


@dataclass(frozen=True)
class BeginSystem(PintParsedStatement):
    """Being of a system directive.

    @system <name> [using <group 1>, ..., <group N>]
    """

    #: Regex to match the header parts of a context.
    _header_re = re.compile(r"@system\s+(?P<name>\w+)\s*(using\s(?P<used_groups>.*))*")

    name: str
    using_group_names: ty.Tuple[str, ...]

    @classmethod
    def from_string(cls, s: str) -> fp.NullableParsedResult[BeginSystem]:
        if not s.startswith("@system"):
            return None

        r = cls._header_re.search(s)

        if r is None:
            raise ValueError("Invalid System header syntax '%s'" % s)

        name = r.groupdict()["name"].strip()
        groups = r.groupdict()["used_groups"]

        # If the systems has no group, it automatically uses the root group.
        if groups:
            group_names = tuple(a.strip() for a in groups.split(","))
        else:
            group_names = ("root",)

        return cls(name, group_names)


@dataclass(frozen=True)
class SystemDefinition(
    block.DirectiveBlock[
        definitions.SystemDefinition,
        BeginSystem,
        ty.Union[plain.CommentDefinition, BaseUnitRule],
    ]
):
    """Definition of a System:

        @system <name> [using <group 1>, ..., <group N>]
            <rule 1>
            ...
            <rule N>
        @end

    See Rule and Comment for more parsing related information.

    The syntax for the rule is:

        new_unit_name : old_unit_name

    where:
        - old_unit_name: a root unit part which is going to be removed from the system.
        - new_unit_name: a non root unit which is going to replace the old_unit.

    If the new_unit_name and the old_unit_name, the later and the colon can be omitted.
    """

    def derive_definition(self) -> definitions.SystemDefinition:
        return definitions.SystemDefinition(
            self.name, self.using_group_names, self.rules
        )

    @property
    def name(self) -> str:
        assert isinstance(self.opening, BeginSystem)
        return self.opening.name

    @property
    def using_group_names(self) -> tuple[str, ...]:
        assert isinstance(self.opening, BeginSystem)
        return self.opening.using_group_names

    @property
    def rules(self) -> tuple[BaseUnitRule, ...]:
        return tuple(el for el in self.body if isinstance(el, BaseUnitRule))



from __future__ import annotations

import copy
import datetime
import logging
import math
import operator as op
import pickle
import warnings
from unittest.mock import patch

import pytest

from pint import (
    DimensionalityError,
    OffsetUnitCalculusError,
    UnitRegistry,
    get_application_registry,
)
from pint.compat import np
from pint.errors import UndefinedBehavior
from pint.facets.plain.unit import UnitsContainer
from pint.testsuite import QuantityTestCase, assert_no_warnings, helpers


class FakeWrapper:
    # Used in test_upcast_type_rejection_on_creation
    def __init__(self, q):
        self.q = q


# TODO: do not subclass from QuantityTestCase
class TestQuantity(QuantityTestCase):
    kwargs = dict(autoconvert_offset_to_baseunit=False)

    def test_quantity_creation(self, caplog):
        for args in (
            (4.2, "meter"),
            (4.2, UnitsContainer(meter=1)),
            (4.2, self.ureg.meter),
            ("4.2*meter",),
            ("4.2/meter**(-1)",),
            (self.Q_(4.2, "meter"),),
        ):
            x = self.Q_(*args)
            assert x.magnitude == 4.2
            assert x.units == UnitsContainer(meter=1)

        x = self.Q_(4.2, UnitsContainer(length=1))
        y = self.Q_(x)
        assert x.magnitude == y.magnitude
        assert x.units == y.units
        assert x is not y

        x = self.Q_(4.2, None)
        assert x.magnitude == 4.2
        assert x.units == UnitsContainer()

        with caplog.at_level(logging.DEBUG):
            assert 4.2 * self.ureg.meter == self.Q_(4.2, 2 * self.ureg.meter)
        assert len(caplog.records) == 1

        assert self.Q_("4.2×10⁻¹² ft/s") == self.Q_(4.2e-12, "foot/second")

    def test_round(self):
        x = self.Q_(1.1, "kg")
        assert isinstance(round(x).magnitude, int)
        assert isinstance(round(x, 0).magnitude, float)

    def test_quantity_with_quantity(self):
        x = self.Q_(4.2, "m")
        assert self.Q_(x, "m").magnitude == 4.2
        assert self.Q_(x, "cm").magnitude == 420.0

    def test_quantity_bool(self):
        assert self.Q_(1, None)
        assert self.Q_(1, "meter")
        assert not self.Q_(0, None)
        assert not self.Q_(0, "meter")
        with pytest.raises(ValueError):
            bool(self.Q_(0, "degC"))
        assert not self.Q_(0, "delta_degC")

    def test_quantity_comparison(self):
        x = self.Q_(4.2, "meter")
        y = self.Q_(4.2, "meter")
        z = self.Q_(5, "meter")
        j = self.Q_(5, "meter*meter")

        # Include a comparison to the application registry
        5 * get_application_registry().meter
        # Include a comparison to a directly created Quantity
        from pint import Quantity

        Quantity(5, "meter")

        # identity for single object
        assert x == x
        assert not (x != x)

        # identity for multiple objects with same value
        assert x == y
        assert not (x != y)

        assert x <= y
        assert x >= y
        assert not (x < y)
        assert not (x > y)

        assert not (x == z)
        assert x != z
        assert x < z

        # TODO: Reinstate this in the near future.
        # Compare with items to the separate application registry
        # assert k >= m  # These should both be from application registry
        # if z._REGISTRY._subregistry != m._REGISTRY._subregistry:
        #     with pytest.raises(ValueError):
        #         z > m  # One from local registry, one from application registry

        assert z != j

        assert z != j
        assert self.Q_(0, "meter") == self.Q_(0, "centimeter")
        assert self.Q_(0, "meter") != self.Q_(0, "second")

        assert self.Q_(10, "meter") < self.Q_(5, "kilometer")

    def test_quantity_comparison_convert(self):
        assert self.Q_(1000, "millimeter") == self.Q_(1, "meter")
        assert self.Q_(1000, "millimeter/min") == self.Q_(1000 / 60, "millimeter/s")

    def test_quantity_repr(self):
        x = self.Q_(4.2, UnitsContainer(meter=1))
        assert str(x) == "4.2 meter"
        assert repr(x) == "<Quantity(4.2, 'meter')>"

    def test_quantity_hash(self):
        x = self.Q_(4.2, "meter")
        x2 = self.Q_(4200, "millimeter")
        y = self.Q_(2, "second")
        z = self.Q_(0.5, "hertz")
        assert hash(x) == hash(x2)

        # Dimensionless equality
        assert hash(y * z) == hash(1.0)

        # Dimensionless equality from a different unit registry
        ureg2 = UnitRegistry(**self.kwargs)
        y2 = ureg2.Quantity(2, "second")
        z2 = ureg2.Quantity(0.5, "hertz")
        assert hash(y * z) == hash(y2 * z2)

    def test_quantity_format(self, subtests):
        x = self.Q_(4.12345678, UnitsContainer(meter=2, kilogram=1, second=-1))
        for spec, result in (
            ("{}", str(x)),
            ("{!s}", str(x)),
            ("{!r}", repr(x)),
            ("{.magnitude}", str(x.magnitude)),
            ("{.units}", str(x.units)),
            ("{.magnitude!s}", str(x.magnitude)),
            ("{.units!s}", str(x.units)),
            ("{.magnitude!r}", repr(x.magnitude)),
            ("{.units!r}", repr(x.units)),
            ("{:.4f}", f"{x.magnitude:.4f} {x.units!s}"),
            (
                "{:L}",
                r"4.12345678\ \frac{\mathrm{kilogram} \cdot \mathrm{meter}^{2}}{\mathrm{second}}",
            ),
            ("{:P}", "4.12345678 kilogram·meter²/second"),
            ("{:H}", "4.12345678 kilogram meter<sup>2</sup>/second"),
            ("{:C}", "4.12345678 kilogram*meter**2/second"),
            ("{:~}", "4.12345678 kg * m ** 2 / s"),
            (
                "{:L~}",
                r"4.12345678\ \frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}",
            ),
            ("{:P~}", "4.12345678 kg·m²/s"),
            ("{:H~}", "4.12345678 kg m<sup>2</sup>/s"),
            ("{:C~}", "4.12345678 kg*m**2/s"),
            ("{:Lx}", r"\SI[]{4.12345678}{\kilo\gram\meter\squared\per\second}"),
        ):
            with subtests.test(spec):
                assert spec.format(x) == result, spec

        # Check the special case that prevents e.g. '3 1 / second'
        x = self.Q_(3, UnitsContainer(second=-1))
        assert f"{x}" == "3 / second"

    @helpers.requires_numpy
    def test_quantity_array_format(self, subtests):
        x = self.Q_(
            np.array([1e-16, 1.0000001, 10000000.0, 1e12, np.nan, np.inf]),
            "kg * m ** 2",
        )
        for spec, result in (
            ("{}", str(x)),
            ("{.magnitude}", str(x.magnitude)),
            (
                "{:e}",
                "[1.000000e-16 1.000000e+00 1.000000e+07 1.000000e+12 nan inf] kilogram * meter ** 2",
            ),
            (
                "{:E}",
                "[1.000000E-16 1.000000E+00 1.000000E+07 1.000000E+12 NAN INF] kilogram * meter ** 2",
            ),
            (
                "{:.2f}",
                "[0.00 1.00 10000000.00 1000000000000.00 nan inf] kilogram * meter ** 2",
            ),
            ("{:.2f~P}", "[0.00 1.00 10000000.00 1000000000000.00 nan inf] kg·m²"),
            ("{:g~P}", "[1e-16 1 1e+07 1e+12 nan inf] kg·m²"),
            (
                "{:.2f~H}",
                (
                    "<table><tbody><tr><th>Magnitude</th><td style='text-align:left;'>"
                    "<pre>[0.00 1.00 10000000.00 1000000000000.00 nan inf]</pre></td></tr>"
                    "<tr><th>Units</th><td style='text-align:left;'>kg m<sup>2</sup></td></tr>"
                    "</tbody></table>"
                ),
            ),
        ):
            with subtests.test(spec):
                assert spec.format(x) == result

    @helpers.requires_numpy
    def test_quantity_array_scalar_format(self, subtests):
        x = self.Q_(np.array(4.12345678), "kg * m ** 2")
        for spec, result in (
            ("{:.2f}", "4.12 kilogram * meter ** 2"),
            ("{:.2fH}", "4.12 kilogram meter<sup>2</sup>"),
        ):
            with subtests.test(spec):
                assert spec.format(x) == result

    def test_format_compact(self):
        q1 = (200e-9 * self.ureg.s).to_compact()
        q1b = self.Q_(200.0, "nanosecond")
        assert round(abs(q1.magnitude - q1b.magnitude), 7) == 0
        assert q1.units == q1b.units

        q2 = (1e-2 * self.ureg("kg m/s^2")).to_compact("N")
        q2b = self.Q_(10.0, "millinewton")
        assert q2.magnitude == q2b.magnitude
        assert q2.units == q2b.units

        q3 = (-1000.0 * self.ureg("meters")).to_compact()
        q3b = self.Q_(-1.0, "kilometer")
        assert q3.magnitude == q3b.magnitude
        assert q3.units == q3b.units

        assert f"{q1:#.1f}" == f"{q1b}"
        assert f"{q2:#.1f}" == f"{q2b}"
        assert f"{q3:#.1f}" == f"{q3b}"

    def test_default_formatting(self, subtests):
        ureg = UnitRegistry()
        x = ureg.Quantity(4.12345678, UnitsContainer(meter=2, kilogram=1, second=-1))
        for spec, result in (
            (
                "L",
                r"4.12345678\ \frac{\mathrm{kilogram} \cdot \mathrm{meter}^{2}}{\mathrm{second}}",
            ),
            ("P", "4.12345678 kilogram·meter²/second"),
            ("H", "4.12345678 kilogram meter<sup>2</sup>/second"),
            ("C", "4.12345678 kilogram*meter**2/second"),
            ("~", "4.12345678 kg * m ** 2 / s"),
            ("L~", r"4.12345678\ \frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}"),
            ("P~", "4.12345678 kg·m²/s"),
            ("H~", "4.12345678 kg m<sup>2</sup>/s"),
            ("C~", "4.12345678 kg*m**2/s"),
        ):
            with subtests.test(spec):
                ureg.formatter.default_format = spec
                assert f"{x}" == result

    def test_formatting_override_default_units(self):
        ureg = UnitRegistry()
        ureg.formatter.default_format = "~"
        x = ureg.Quantity(4, "m ** 2")

        assert f"{x:dP}" == "4 meter²"
        ureg.separate_format_defaults = None
        with pytest.warns(DeprecationWarning):
            assert f"{x:d}" == "4 meter ** 2"

        ureg.separate_format_defaults = True
        with assert_no_warnings():
            assert f"{x:d}" == "4 m ** 2"

    def test_formatting_override_default_magnitude(self):
        ureg = UnitRegistry()
        ureg.formatter.default_format = ".2f"
        x = ureg.Quantity(4, "m ** 2")

        assert f"{x:dP}" == "4 meter²"
        ureg.separate_format_defaults = None
        with pytest.warns(DeprecationWarning):
            assert f"{x:D}" == "4 meter ** 2"

        ureg.separate_format_defaults = True
        with assert_no_warnings():
            assert f"{x:D}" == "4.00 meter ** 2"

    def test_exponent_formatting(self):
        ureg = UnitRegistry()
        x = ureg.Quantity(1e20, "meter")
        assert f"{x:~H}" == r"1×10<sup>20</sup> m"
        assert f"{x:~L}" == r"1\times 10^{20}\ \mathrm{m}"
        assert f"{x:~Lx}" == r"\SI[]{1e+20}{\meter}"
        assert f"{x:~P}" == r"1×10²⁰ m"

        x = ureg.Quantity(1e-20, "meter")
        assert f"{x:~H}" == r"1×10<sup>-20</sup> m"
        assert f"{x:~L}" == r"1\times 10^{-20}\ \mathrm{m}"
        assert f"{x:~Lx}" == r"\SI[]{1e-20}{\meter}"
        assert f"{x:~P}" == r"1×10⁻²⁰ m"

    def test_ipython(self):
        alltext = []

        class Pretty:
            @staticmethod
            def text(text):
                alltext.append(text)

            @classmethod
            def pretty(cls, data):
                try:
                    data._repr_pretty_(cls, False)
                except AttributeError:
                    alltext.append(str(data))

        ureg = UnitRegistry()
        x = 3.5 * ureg.Unit(UnitsContainer(meter=2, kilogram=1, second=-1))
        assert x._repr_html_() == "3.5 kilogram meter<sup>2</sup>/second"
        assert (
            x._repr_latex_() == r"$3.5\ \frac{\mathrm{kilogram} \cdot "
            r"\mathrm{meter}^{2}}{\mathrm{second}}$"
        )
        x._repr_pretty_(Pretty, False)
        assert "".join(alltext) == "3.5 kilogram·meter²/second"
        ureg.formatter.default_format = "~"
        assert x._repr_html_() == "3.5 kg m<sup>2</sup>/s"
        assert (
            x._repr_latex_() == r"$3.5\ \frac{\mathrm{kg} \cdot "
            r"\mathrm{m}^{2}}{\mathrm{s}}$"
        )
        alltext = []
        x._repr_pretty_(Pretty, False)
        assert "".join(alltext) == "3.5 kg·m²/s"

    def test_to_base_units(self):
        x = self.Q_("1*inch")
        helpers.assert_quantity_almost_equal(
            x.to_base_units(), self.Q_(0.0254, "meter")
        )
        x = self.Q_("1*inch*inch")
        helpers.assert_quantity_almost_equal(
            x.to_base_units(), self.Q_(0.0254**2.0, "meter*meter")
        )
        x = self.Q_("1*inch/minute")
        helpers.assert_quantity_almost_equal(
            x.to_base_units(), self.Q_(0.0254 / 60.0, "meter/second")
        )

    def test_convert(self):
        helpers.assert_quantity_almost_equal(
            self.Q_("2 inch").to("meter"), self.Q_(2.0 * 0.0254, "meter")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_("2 meter").to("inch"), self.Q_(2.0 / 0.0254, "inch")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_("2 sidereal_year").to("second"), self.Q_(63116297.5325, "second")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_("2.54 centimeter/second").to("inch/second"),
            self.Q_("1 inch/second"),
        )
        assert round(abs(self.Q_("2.54 centimeter").to("inch").magnitude - 1), 7) == 0
        assert (
            round(abs(self.Q_("2 second").to("millisecond").magnitude - 2000), 7) == 0
        )

    @helpers.requires_mip
    def test_to_preferred(self):
        ureg = self.ureg
        Q_ = self.Q_

        ureg.define("pound_force_per_square_foot = 47.8803 pascals = psf")
        ureg.define("pound_mass = 0.45359237 kg = lbm")

        preferred_units = [
            ureg.ft,  # distance      L
            ureg.slug,  # mass          M
            ureg.s,  # duration      T
            ureg.rankine,  # temperature   Θ
            ureg.lbf,  # force         L M T^-2
            ureg.psf,  # pressure      M L^−1 T^−2
            ureg.lbm * ureg.ft**-3,  # density       M L^-3
            ureg.W,  # power         L^2 M T^-3
        ]

        temp = (Q_("1 lbf") * Q_("1 m/s")).to_preferred(preferred_units)
        assert temp.units == ureg.W

        temp = (Q_(" 1 lbf*m")).to_preferred(preferred_units)
        # would prefer this to be repeatable, but mip doesn't guarantee that currently
        assert temp.units in (ureg.W * ureg.s, ureg.ft * ureg.lbf)

        temp = Q_("1 kg").to_preferred(preferred_units)
        assert temp.units == ureg.slug

        result = Q_("1 slug/m**3").to_preferred(preferred_units)
        assert result.units == ureg.lbm * ureg.ft**-3

        result = Q_("1 amp").to_preferred(preferred_units)
        assert result.units == ureg.amp

        result = Q_("1 volt").to_preferred(preferred_units)
        assert result.units == ureg.volts

    @helpers.requires_mip
    def test_to_preferred_registry(self):
        ureg = self.ureg
        Q_ = self.Q_
        ureg.default_preferred_units = [
            ureg.m,  # distance      L
            ureg.kg,  # mass          M
            ureg.s,  # duration      T
            ureg.N,  # force         L M T^-2
            ureg.Pa,  # pressure      M L^−1 T^−2
            ureg.W,  # power         L^2 M T^-3
        ]
        pressure = (Q_(1, "N") * Q_("1 m**-2")).to_preferred()
        assert pressure.units == ureg.Pa

    @helpers.requires_mip
    def test_autoconvert_to_preferred(self):
        ureg = self.ureg
        Q_ = self.Q_
        ureg.autoconvert_to_preferred = True
        ureg.default_preferred_units = [
            ureg.m,  # distance      L
            ureg.kg,  # mass          M
            ureg.s,  # duration      T
            ureg.N,  # force         L M T^-2
            ureg.Pa,  # pressure      M L^−1 T^−2
            ureg.W,  # power         L^2 M T^-3
        ]
        pressure = Q_(1, "N") * Q_("1 m**-2")
        assert pressure.units == ureg.Pa

    @helpers.requires_numpy
    def test_convert_numpy(self):
        # Conversions with single units take a different codepath than
        # Conversions with more than one unit.
        src_dst1 = UnitsContainer(meter=1), UnitsContainer(inch=1)
        src_dst2 = UnitsContainer(meter=1, second=-1), UnitsContainer(inch=1, minute=-1)
        for src, dst in (src_dst1, src_dst2):
            a = np.ones((3, 1))
            ac = np.ones((3, 1))

            q = self.Q_(a, src)
            qac = self.Q_(ac, src).to(dst)
            r = q.to(dst)
            helpers.assert_quantity_almost_equal(qac, r)
            assert r is not q
            assert r._magnitude is not a

    def test_convert_from(self):
        x = self.Q_("2*inch")
        meter = self.ureg.meter

        # from quantity
        helpers.assert_quantity_almost_equal(
            meter.from_(x), self.Q_(2.0 * 0.0254, "meter")
        )
        helpers.assert_quantity_almost_equal(meter.m_from(x), 2.0 * 0.0254)

        # from unit
        helpers.assert_quantity_almost_equal(
            meter.from_(self.ureg.inch), self.Q_(0.0254, "meter")
        )
        helpers.assert_quantity_almost_equal(meter.m_from(self.ureg.inch), 0.0254)

        # from number
        helpers.assert_quantity_almost_equal(
            meter.from_(2, strict=False), self.Q_(2.0, "meter")
        )
        helpers.assert_quantity_almost_equal(meter.m_from(2, strict=False), 2.0)

        # from number (strict mode)
        with pytest.raises(ValueError):
            meter.from_(2)
        with pytest.raises(ValueError):
            meter.m_from(2)

    @helpers.requires_numpy
    def test_retain_unit(self):
        # Test that methods correctly retain units and do not degrade into
        # ordinary ndarrays.  List contained in __copy_units.
        a = np.ones((3, 2))
        q = self.Q_(a, "km")
        assert q.u == q.reshape(2, 3).u
        assert q.u == q.swapaxes(0, 1).u
        assert q.u == q.mean().u
        assert q.u == np.compress((q == q[0, 0]).any(0), q).u

    def test_context_attr(self):
        assert self.ureg.meter == self.Q_(1, "meter")

    def test_both_symbol(self):
        assert self.Q_(2, "ms") == self.Q_(2, "millisecond")
        assert self.Q_(2, "cm") == self.Q_(2, "centimeter")
        assert self.Q_(2, "mm / s ** 2") == self.Q_(2, "millimeter_per_second_squared")

    def test_dimensionless_units(self):
        assert (
            round(abs(self.Q_(360, "degree").to("radian").magnitude - 2 * math.pi), 7)
            == 0
        )
        assert (
            round(abs(self.Q_(2 * math.pi, "radian") - self.Q_(360, "degree")), 7) == 0
        )
        assert self.Q_(1, "radian").dimensionality == UnitsContainer()
        assert self.Q_(1, "radian").dimensionless
        assert not self.Q_(1, "radian").unitless

        assert self.Q_(1, "meter") / self.Q_(1, "meter") == 1
        assert (self.Q_(1, "meter") / self.Q_(1, "mm")).to("") == 1000

        assert self.Q_(10) // self.Q_(360, "degree") == 1
        assert self.Q_(400, "degree") // self.Q_(2 * math.pi) == 1
        assert self.Q_(400, "degree") // (2 * math.pi) == 1
        assert 7 // self.Q_(360, "degree") == 1

    def test_offset(self):
        helpers.assert_quantity_almost_equal(
            self.Q_(0, "kelvin").to("kelvin"), self.Q_(0, "kelvin")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(0, "degC").to("kelvin"), self.Q_(273.15, "kelvin")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(0, "degF").to("kelvin"), self.Q_(255.372222, "kelvin"), rtol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(100, "kelvin").to("kelvin"), self.Q_(100, "kelvin")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "degC").to("kelvin"), self.Q_(373.15, "kelvin")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "degF").to("kelvin"),
            self.Q_(310.92777777, "kelvin"),
            rtol=0.01,
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(0, "kelvin").to("degC"), self.Q_(-273.15, "degC")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "kelvin").to("degC"), self.Q_(-173.15, "degC")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(0, "kelvin").to("degF"), self.Q_(-459.67, "degF"), rtol=0.01
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "kelvin").to("degF"), self.Q_(-279.67, "degF"), rtol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(32, "degF").to("degC"), self.Q_(0, "degC"), atol=0.01
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "degC").to("degF"), self.Q_(212, "degF"), atol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(54, "degF").to("degC"), self.Q_(12.2222, "degC"), atol=0.01
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(12, "degC").to("degF"), self.Q_(53.6, "degF"), atol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(12, "kelvin").to("degC"), self.Q_(-261.15, "degC"), atol=0.01
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(12, "degC").to("kelvin"), self.Q_(285.15, "kelvin"), atol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(12, "kelvin").to("degR"), self.Q_(21.6, "degR"), atol=0.01
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(12, "degR").to("kelvin"), self.Q_(6.66666667, "kelvin"), atol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(12, "degC").to("degR"), self.Q_(513.27, "degR"), atol=0.01
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(12, "degR").to("degC"), self.Q_(-266.483333, "degC"), atol=0.01
        )

    def test_offset_delta(self):
        helpers.assert_quantity_almost_equal(
            self.Q_(0, "delta_degC").to("kelvin"), self.Q_(0, "kelvin")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(0, "delta_degF").to("kelvin"), self.Q_(0, "kelvin"), rtol=0.01
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(100, "kelvin").to("delta_degC"), self.Q_(100, "delta_degC")
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "kelvin").to("delta_degF"),
            self.Q_(180, "delta_degF"),
            rtol=0.01,
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "delta_degF").to("kelvin"),
            self.Q_(55.55555556, "kelvin"),
            rtol=0.01,
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "delta_degC").to("delta_degF"),
            self.Q_(180, "delta_degF"),
            rtol=0.01,
        )
        helpers.assert_quantity_almost_equal(
            self.Q_(100, "delta_degF").to("delta_degC"),
            self.Q_(55.55555556, "delta_degC"),
            rtol=0.01,
        )

        helpers.assert_quantity_almost_equal(
            self.Q_(12.3, "delta_degC").to("delta_degF"),
            self.Q_(22.14, "delta_degF"),
            rtol=0.01,
        )

    def test_pickle(self, subtests):
        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            for magnitude, unit in ((32, ""), (2.4, ""), (32, "m/s"), (2.4, "m/s")):
                with subtests.test(protocol=protocol, magnitude=magnitude, unit=unit):
                    q1 = self.Q_(magnitude, unit)
                    q2 = pickle.loads(pickle.dumps(q1, protocol))
                    assert q1 == q2

    @helpers.requires_numpy
    def test_from_sequence(self):
        u_array_ref = self.Q_([200, 1000], "g")
        u_array_ref_reversed = self.Q_([1000, 200], "g")
        u_seq = [self.Q_("200g"), self.Q_("1kg")]
        u_seq_reversed = u_seq[::-1]

        u_array = self.Q_.from_sequence(u_seq)
        assert all(u_array == u_array_ref)

        u_array_2 = self.Q_.from_sequence(u_seq_reversed)
        assert all(u_array_2 == u_array_ref_reversed)
        assert not (u_array_2.u == u_array_ref_reversed.u)

        u_array_3 = self.Q_.from_sequence(u_seq_reversed, units="g")
        assert all(u_array_3 == u_array_ref_reversed)
        assert u_array_3.u == u_array_ref_reversed.u

        with pytest.raises(ValueError):
            self.Q_.from_sequence([])

        u_array_5 = self.Q_.from_list(u_seq)
        assert all(u_array_5 == u_array_ref)

    @helpers.requires_numpy
    def test_iter(self):
        # Verify that iteration gives element as Quantity with same units
        x = self.Q_([0, 1, 2, 3], "m")
        helpers.assert_quantity_equal(next(iter(x)), self.Q_(0, "m"))

    def test_notiter(self):
        # Verify that iter() crashes immediately, without needing to draw any
        # element from it, if the magnitude isn't iterable
        x = self.Q_(1, "m")
        with pytest.raises(TypeError):
            iter(x)

    @helpers.requires_array_function_protocol()
    def test_no_longer_array_function_warning_on_creation(self):
        # Test that warning is no longer raised on first creation
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            self.Q_([])

    @helpers.requires_not_numpy()
    def test_no_ndarray_coercion_without_numpy(self):
        with pytest.raises(ValueError):
            self.Q_(1, "m").__array__()

    @patch(
        "pint.compat.upcast_type_names", ("pint.testsuite.test_quantity.FakeWrapper",)
    )
    @patch(
        "pint.compat.upcast_type_map",
        {"pint.testsuite.test_quantity.FakeWrapper": FakeWrapper},
    )
    def test_upcast_type_rejection_on_creation(self):
        with pytest.raises(TypeError):
            self.Q_(FakeWrapper(42), "m")
        assert FakeWrapper(self.Q_(42, "m")).q == self.Q_(42, "m")

    def test_is_compatible_with(self):
        a = self.Q_(1, "kg")
        b = self.Q_(20, "g")
        c = self.Q_(550)

        assert a.is_compatible_with(b)
        assert a.is_compatible_with("lb")
        assert a.is_compatible_with(self.U_("lb"))
        assert not a.is_compatible_with("km")
        assert not a.is_compatible_with("")
        assert not a.is_compatible_with(12)

        assert c.is_compatible_with(12)

    def test_is_compatible_with_with_context(self):
        a = self.Q_(532.0, "nm")
        b = self.Q_(563.5, "terahertz")
        assert a.is_compatible_with(b, "sp")
        with self.ureg.context("sp"):
            assert a.is_compatible_with(b)

    @pytest.mark.parametrize(["inf_str"], [("inf",), ("-infinity",), ("INFINITY",)])
    @pytest.mark.parametrize(["has_unit"], [(True,), (False,)])
    def test_infinity(self, inf_str, has_unit):
        inf = float(inf_str)
        ref = self.Q_(inf, "meter" if has_unit else None)
        test = self.Q_(inf_str + (" meter" if has_unit else ""))
        assert ref == test

    @pytest.mark.parametrize(["nan_str"], [("nan",), ("NAN",)])
    @pytest.mark.parametrize(["has_unit"], [(True,), (False,)])
    def test_nan(self, nan_str, has_unit):
        nan = float(nan_str)
        ref = self.Q_(nan, " meter" if has_unit else None)
        test = self.Q_(nan_str + (" meter" if has_unit else ""))
        assert ref.units == test.units
        assert math.isnan(test.magnitude)
        assert ref != test

    @helpers.requires_numpy
    def test_to_reduced_units(self):
        q = self.Q_([3, 4], "s * ms")
        helpers.assert_quantity_equal(
            q.to_reduced_units(), self.Q_([3000.0, 4000.0], "ms**2")
        )

        q = self.Q_(0.5, "g*t/kg")
        helpers.assert_quantity_equal(q.to_reduced_units(), self.Q_(0.5, "kg"))

    def test_to_reduced_units_dimensionless(self):
        ureg = UnitRegistry(preprocessors=[lambda x: x.replace("%", " percent ")])
        ureg.define("percent = 0.01 count = %")
        Q_ = ureg.Quantity
        reduced_quantity = (Q_("1 s") * Q_("5 %") / Q_("1 count")).to_reduced_units()
        assert reduced_quantity == ureg.Quantity(0.05, ureg.second)

    @pytest.mark.parametrize(
        ("unit_str", "expected_unit"),
        [
            ("hour/hr", {}),
            ("cm centimeter cm centimeter", {"centimeter": 4}),
        ],
    )
    def test_unit_canonical_name_parsing(self, unit_str, expected_unit):
        q = self.Q_(1, unit_str)
        assert q._units == UnitsContainer(expected_unit)


# TODO: do not subclass from QuantityTestCase
class TestQuantityToCompact(QuantityTestCase):
    def assertQuantityAlmostIdentical(self, q1, q2):
        assert q1.units == q2.units
        assert round(abs(q1.magnitude - q2.magnitude), 7) == 0

    def compare_quantity_compact(self, q, expected_compact, unit=None):
        helpers.assert_quantity_almost_equal(q.to_compact(unit=unit), expected_compact)

    def test_dimensionally_simple_units(self):
        ureg = self.ureg
        self.compare_quantity_compact(1 * ureg.m, 1 * ureg.m)
        self.compare_quantity_compact(1e-9 * ureg.m, 1 * ureg.nm)

    def test_power_units(self):
        ureg = self.ureg
        self.compare_quantity_compact(900 * ureg.m**2, 900 * ureg.m**2)
        self.compare_quantity_compact(1e7 * ureg.m**2, 10 * ureg.km**2)

    def test_inverse_units(self):
        ureg = self.ureg
        self.compare_quantity_compact(1 / ureg.m, 1 / ureg.m)
        self.compare_quantity_compact(100e9 / ureg.m, 100 / ureg.nm)

    def test_inverse_square_units(self):
        ureg = self.ureg
        self.compare_quantity_compact(1 / ureg.m**2, 1 / ureg.m**2)
        self.compare_quantity_compact(1e11 / ureg.m**2, 1e5 / ureg.mm**2)

    def test_fractional_units(self):
        ureg = self.ureg
        # Typing denominator first to provoke potential error
        self.compare_quantity_compact(20e3 * ureg("hr^(-1) m"), 20 * ureg.km / ureg.hr)

    def test_fractional_exponent_units(self):
        ureg = self.ureg
        self.compare_quantity_compact(1 * ureg.m**0.5, 1 * ureg.m**0.5)
        self.compare_quantity_compact(1e-2 * ureg.m**0.5, 10 * ureg.um**0.5)

    def test_derived_units(self):
        ureg = self.ureg
        self.compare_quantity_compact(0.5 * ureg.megabyte, 500 * ureg.kilobyte)
        self.compare_quantity_compact(1e-11 * ureg.N, 10 * ureg.pN)

    def test_unit_parameter(self):
        ureg = self.ureg
        self.compare_quantity_compact(
            self.Q_(100e-9, "kg m / s^2"), 100 * ureg.nN, ureg.N
        )
        self.compare_quantity_compact(
            self.Q_(101.3e3, "kg/m/s^2"), 101.3 * ureg.kPa, ureg.Pa
        )

    def test_limits_magnitudes(self):
        ureg = self.ureg
        self.compare_quantity_compact(0 * ureg.m, 0 * ureg.m)
        self.compare_quantity_compact(float("inf") * ureg.m, float("inf") * ureg.m)

    def test_nonnumeric_magnitudes(self):
        ureg = self.ureg
        x = "some string" * ureg.m
        with pytest.warns(UndefinedBehavior):
            x.to_compact()

    def test_very_large_to_compact(self):
        # This should not raise an IndexError
        self.compare_quantity_compact(
            self.Q_(10000, "yottameter"), self.Q_(10**28, "meter").to_compact()
        )


# TODO: do not subclass from QuantityTestCase
class TestQuantityBasicMath(QuantityTestCase):
    def _test_inplace(self, operator, value1, value2, expected_result, unit=None):
        if isinstance(value1, str):
            value1 = self.Q_(value1)
        if isinstance(value2, str):
            value2 = self.Q_(value2)
        if isinstance(expected_result, str):
            expected_result = self.Q_(expected_result)

        if unit is not None:
            value1 = value1 * unit
            value2 = value2 * unit
            expected_result = expected_result * unit

        value1 = copy.copy(value1)
        value2 = copy.copy(value2)
        id1 = id(value1)
        id2 = id(value2)
        value1 = operator(value1, value2)
        value2_cpy = copy.copy(value2)
        helpers.assert_quantity_almost_equal(value1, expected_result)
        assert id1 == id(value1)
        helpers.assert_quantity_almost_equal(value2, value2_cpy)
        assert id2 == id(value2)

    def _test_not_inplace(self, operator, value1, value2, expected_result, unit=None):
        if isinstance(value1, str):
            value1 = self.Q_(value1)
        if isinstance(value2, str):
            value2 = self.Q_(value2)
        if isinstance(expected_result, str):
            expected_result = self.Q_(expected_result)

        if unit is not None:
            value1 = value1 * unit
            value2 = value2 * unit
            expected_result = expected_result * unit

        id1 = id(value1)
        id2 = id(value2)

        value1_cpy = copy.copy(value1)
        value2_cpy = copy.copy(value2)

        result = operator(value1, value2)

        helpers.assert_quantity_almost_equal(expected_result, result)
        helpers.assert_quantity_almost_equal(value1, value1_cpy)
        helpers.assert_quantity_almost_equal(value2, value2_cpy)
        assert id(result) != id1
        assert id(result) != id2

    def _test_quantity_add_sub(self, unit, func):
        x = self.Q_(unit, "centimeter")
        y = self.Q_(unit, "inch")
        z = self.Q_(unit, "second")
        a = self.Q_(unit, None)

        func(op.add, x, x, self.Q_(unit + unit, "centimeter"))
        func(op.add, x, y, self.Q_(unit + 2.54 * unit, "centimeter"))
        func(op.add, y, x, self.Q_(unit + unit / (2.54 * unit), "inch"))
        func(op.add, a, unit, self.Q_(unit + unit, None))
        with pytest.raises(DimensionalityError):
            op.add(10, x)
        with pytest.raises(DimensionalityError):
            op.add(x, 10)
        with pytest.raises(DimensionalityError):
            op.add(x, z)

        func(op.sub, x, x, self.Q_(unit - unit, "centimeter"))
        func(op.sub, x, y, self.Q_(unit - 2.54 * unit, "centimeter"))
        func(op.sub, y, x, self.Q_(unit - unit / (2.54 * unit), "inch"))
        func(op.sub, a, unit, self.Q_(unit - unit, None))
        with pytest.raises(DimensionalityError):
            op.sub(10, x)
        with pytest.raises(DimensionalityError):
            op.sub(x, 10)
        with pytest.raises(DimensionalityError):
            op.sub(x, z)

    def _test_quantity_iadd_isub(self, unit, func):
        x = self.Q_(unit, "centimeter")
        y = self.Q_(unit, "inch")
        z = self.Q_(unit, "second")
        a = self.Q_(unit, None)

        func(op.iadd, x, x, self.Q_(unit + unit, "centimeter"))
        func(op.iadd, x, y, self.Q_(unit + 2.54 * unit, "centimeter"))
        func(op.iadd, y, x, self.Q_(unit + unit / 2.54, "inch"))
        func(op.iadd, a, unit, self.Q_(unit + unit, None))
        with pytest.raises(DimensionalityError):
            op.iadd(10, x)
        with pytest.raises(DimensionalityError):
            op.iadd(x, 10)
        with pytest.raises(DimensionalityError):
            op.iadd(x, z)

        func(op.isub, x, x, self.Q_(unit - unit, "centimeter"))
        func(op.isub, x, y, self.Q_(unit - 2.54, "centimeter"))
        func(op.isub, y, x, self.Q_(unit - unit / 2.54, "inch"))
        func(op.isub, a, unit, self.Q_(unit - unit, None))
        with pytest.raises(DimensionalityError):
            op.sub(10, x)
        with pytest.raises(DimensionalityError):
            op.sub(x, 10)
        with pytest.raises(DimensionalityError):
            op.sub(x, z)

    def _test_quantity_mul_div(self, unit, func):
        func(op.mul, unit * 10.0, "4.2*meter", "42*meter", unit)
        func(op.mul, "4.2*meter", unit * 10.0, "42*meter", unit)
        func(op.mul, "4.2*meter", "10*inch", "42*meter*inch", unit)
        func(op.truediv, unit * 42, "4.2*meter", "10/meter", unit)
        func(op.truediv, "4.2*meter", unit * 10.0, "0.42*meter", unit)
        func(op.truediv, "4.2*meter", "10*inch", "0.42*meter/inch", unit)

    def _test_quantity_imul_idiv(self, unit, func):
        # func(op.imul, 10.0, '4.2*meter', '42*meter')
        func(op.imul, "4.2*meter", 10.0, "42*meter", unit)
        func(op.imul, "4.2*meter", "10*inch", "42*meter*inch", unit)
        # func(op.truediv, 42, '4.2*meter', '10/meter')
        func(op.itruediv, "4.2*meter", unit * 10.0, "0.42*meter", unit)
        func(op.itruediv, "4.2*meter", "10*inch", "0.42*meter/inch", unit)

    def _test_quantity_floordiv(self, unit, func):
        a = self.Q_("10*meter")
        b = self.Q_("3*second")
        with pytest.raises(DimensionalityError):
            op.floordiv(a, b)
        with pytest.raises(DimensionalityError):
            op.floordiv(3, b)
        with pytest.raises(DimensionalityError):
            op.floordiv(a, 3)
        with pytest.raises(DimensionalityError):
            op.ifloordiv(a, b)
        with pytest.raises(DimensionalityError):
            op.ifloordiv(3, b)
        with pytest.raises(DimensionalityError):
            op.ifloordiv(a, 3)
        func(op.floordiv, unit * 10.0, "4.2*meter/meter", 2, unit)
        func(op.floordiv, "10*meter", "4.2*inch", 93, unit)

    def _test_quantity_mod(self, unit, func):
        a = self.Q_("10*meter")
        b = self.Q_("3*second")
        with pytest.raises(DimensionalityError):
            op.mod(a, b)
        with pytest.raises(DimensionalityError):
            op.mod(3, b)
        with pytest.raises(DimensionalityError):
            op.mod(a, 3)
        with pytest.raises(DimensionalityError):
            op.imod(a, b)
        with pytest.raises(DimensionalityError):
            op.imod(3, b)
        with pytest.raises(DimensionalityError):
            op.imod(a, 3)
        func(op.mod, unit * 10.0, "4.2*meter/meter", 1.6, unit)

    def _test_quantity_ifloordiv(self, unit, func):
        func(op.ifloordiv, 10.0, "4.2*meter/meter", 2, unit)
        func(op.ifloordiv, "10*meter", "4.2*inch", 93, unit)

    def _test_quantity_divmod_one(self, a, b):
        if isinstance(a, str):
            a = self.Q_(a)
        if isinstance(b, str):
            b = self.Q_(b)

        q, r = divmod(a, b)
        assert q == a // b
        assert r == a % b
        assert a == (q * b) + r
        assert q == math.floor(q)
        if b > (0 * b):
            assert (0 * b) <= r < b
        else:
            assert (0 * b) >= r > b
        if isinstance(a, self.Q_):
            assert r.units == a.units
        else:
            assert r.unitless
        assert q.unitless

        copy_a = copy.copy(a)
        a %= b
        assert a == r
        copy_a //= b
        assert copy_a == q

    def _test_quantity_divmod(self):
        self._test_quantity_divmod_one("10*meter", "4.2*inch")
        self._test_quantity_divmod_one("-10*meter", "4.2*inch")
        self._test_quantity_divmod_one("-10*meter", "-4.2*inch")
        self._test_quantity_divmod_one("10*meter", "-4.2*inch")

        self._test_quantity_divmod_one("400*degree", "3")
        self._test_quantity_divmod_one("4", "180 degree")
        self._test_quantity_divmod_one(4, "180 degree")
        self._test_quantity_divmod_one("20", 4)
        self._test_quantity_divmod_one("300*degree", "100 degree")

        a = self.Q_("10*meter")
        b = self.Q_("3*second")
        with pytest.raises(DimensionalityError):
            divmod(a, b)
        with pytest.raises(DimensionalityError):
            divmod(3, b)
        with pytest.raises(DimensionalityError):
            divmod(a, 3)

    def _test_numeric(self, unit, ifunc):
        self._test_quantity_add_sub(unit, self._test_not_inplace)
        self._test_quantity_iadd_isub(unit, ifunc)
        self._test_quantity_mul_div(unit, self._test_not_inplace)
        self._test_quantity_imul_idiv(unit, ifunc)
        self._test_quantity_floordiv(unit, self._test_not_inplace)
        self._test_quantity_mod(unit, self._test_not_inplace)
        self._test_quantity_divmod()
        # self._test_quantity_ifloordiv(unit, ifunc)

    def test_float(self):
        self._test_numeric(1.0, self._test_not_inplace)

    def test_fraction(self):
        import fractions

        self._test_numeric(fractions.Fraction(1, 1), self._test_not_inplace)

    @helpers.requires_numpy
    def test_nparray(self):
        self._test_numeric(np.ones((1, 3)), self._test_inplace)

    def test_quantity_abs_round(self):
        x = self.Q_(-4.2, "meter")
        y = self.Q_(4.2, "meter")

        for fun in (abs, round, op.pos, op.neg):
            zx = self.Q_(fun(x.magnitude), "meter")
            zy = self.Q_(fun(y.magnitude), "meter")
            rx = fun(x)
            ry = fun(y)
            assert rx == zx, f"while testing {fun}"
            assert ry == zy, f"while testing {fun}"
            assert rx is not zx, f"while testing {fun}"
            assert ry is not zy, f"while testing {fun}"

    def test_quantity_float_complex(self):
        x = self.Q_(-4.2, None)
        y = self.Q_(4.2, None)
        z = self.Q_(1, "meter")
        for fun in (float, complex):
            assert fun(x) == fun(x.magnitude)
            assert fun(y) == fun(y.magnitude)
            with pytest.raises(DimensionalityError):
                fun(z)


# TODO: do not subclass from QuantityTestCase
class TestQuantityNeutralAdd(QuantityTestCase):
    """Addition to zero or NaN is allowed between a Quantity and a non-Quantity"""

    def test_bare_zero(self):
        v = self.Q_(2.0, "m")
        assert v + 0 == v
        assert v - 0 == v
        assert 0 + v == v
        assert 0 - v == -v

    def test_bare_zero_inplace(self):
        v = self.Q_(2.0, "m")
        v2 = self.Q_(2.0, "m")
        v2 += 0
        assert v2 == v
        v2 = self.Q_(2.0, "m")
        v2 -= 0
        assert v2 == v
        v2 = 0
        v2 += v
        assert v2 == v
        v2 = 0
        v2 -= v
        assert v2 == -v

    def test_bare_nan(self):
        v = self.Q_(2.0, "m")
        helpers.assert_quantity_equal(v + math.nan, self.Q_(math.nan, v.units))
        helpers.assert_quantity_equal(v - math.nan, self.Q_(math.nan, v.units))
        helpers.assert_quantity_equal(math.nan + v, self.Q_(math.nan, v.units))
        helpers.assert_quantity_equal(math.nan - v, self.Q_(math.nan, v.units))

    def test_bare_nan_inplace(self):
        v = self.Q_(2.0, "m")
        v2 = self.Q_(2.0, "m")
        v2 += math.nan
        helpers.assert_quantity_equal(v2, self.Q_(math.nan, v.units))
        v2 = self.Q_(2.0, "m")
        v2 -= math.nan
        helpers.assert_quantity_equal(v2, self.Q_(math.nan, v.units))
        v2 = math.nan
        v2 += v
        helpers.assert_quantity_equal(v2, self.Q_(math.nan, v.units))
        v2 = math.nan
        v2 -= v
        helpers.assert_quantity_equal(v2, self.Q_(math.nan, v.units))

    @helpers.requires_numpy
    def test_bare_zero_or_nan_numpy(self):
        z = np.array([0.0, np.nan])
        v = self.Q_([1.0, 2.0], "m")
        e = self.Q_([1.0, np.nan], "m")
        helpers.assert_quantity_equal(z + v, e)
        helpers.assert_quantity_equal(z - v, -e)
        helpers.assert_quantity_equal(v + z, e)
        helpers.assert_quantity_equal(v - z, e)

        # If any element is non-zero and non-NaN, raise DimensionalityError
        nz = np.array([0.0, 1.0])
        with pytest.raises(DimensionalityError):
            nz + v
        with pytest.raises(DimensionalityError):
            nz - v
        with pytest.raises(DimensionalityError):
            v + nz
        with pytest.raises(DimensionalityError):
            v - nz

        # Mismatched shape
        z = np.array([0.0, np.nan, 0.0])
        v = self.Q_([1.0, 2.0], "m")
        for x, y in ((z, v), (v, z)):
            with pytest.raises(ValueError):
                x + y
            with pytest.raises(ValueError):
                x - y

    @helpers.requires_numpy
    def test_bare_zero_or_nan_numpy_inplace(self):
        z = np.array([0.0, np.nan])
        v = self.Q_([1.0, 2.0], "m")
        e = self.Q_([1.0, np.nan], "m")
        v += z
        helpers.assert_quantity_equal(v, e)
        v = self.Q_([1.0, 2.0], "m")
        v -= z
        helpers.assert_quantity_equal(v, e)
        v = self.Q_([1.0, 2.0], "m")
        z = np.array([0.0, np.nan])
        z += v
        helpers.assert_quantity_equal(z, e)
        v = self.Q_([1.0, 2.0], "m")
        z = np.array([0.0, np.nan])
        z -= v
        helpers.assert_quantity_equal(z, -e)


# TODO: do not subclass from QuantityTestCase
class TestDimensions(QuantityTestCase):
    def test_get_dimensionality(self):
        get = self.ureg.get_dimensionality
        assert get("[time]") == UnitsContainer({"[time]": 1})
        assert get(UnitsContainer({"[time]": 1})) == UnitsContainer({"[time]": 1})
        assert get("seconds") == UnitsContainer({"[time]": 1})
        assert get(UnitsContainer({"seconds": 1})) == UnitsContainer({"[time]": 1})
        assert get("[velocity]") == UnitsContainer({"[length]": 1, "[time]": -1})
        assert get("[acceleration]") == UnitsContainer({"[length]": 1, "[time]": -2})

    def test_dimensionality(self):
        x = self.Q_(42, "centimeter")
        x.to_base_units()
        x = self.Q_(42, "meter*second")
        assert x.dimensionality == UnitsContainer({"[length]": 1.0, "[time]": 1.0})
        x = self.Q_(42, "meter*second*second")
        assert x.dimensionality == UnitsContainer({"[length]": 1.0, "[time]": 2.0})
        x = self.Q_(42, "inch*second*second")
        assert x.dimensionality == UnitsContainer({"[length]": 1.0, "[time]": 2.0})
        assert self.Q_(42, None).dimensionless
        assert not self.Q_(42, "meter").dimensionless
        assert (self.Q_(42, "meter") / self.Q_(1, "meter")).dimensionless
        assert not (self.Q_(42, "meter") / self.Q_(1, "second")).dimensionless
        assert (self.Q_(42, "meter") / self.Q_(1, "inch")).dimensionless

    def test_inclusion(self):
        dim = self.Q_(42, "meter").dimensionality
        assert "[length]" in dim
        assert "[time]" not in dim
        dim = (self.Q_(42, "meter") / self.Q_(11, "second")).dimensionality
        assert "[length]" in dim
        assert "[time]" in dim
        dim = self.Q_(20.785, "J/(mol)").dimensionality
        for dimension in ("[length]", "[mass]", "[substance]", "[time]"):
            assert dimension in dim
        assert "[angle]" not in dim


class TestQuantityWithDefaultRegistry(TestQuantity):
    @classmethod
    def setup_class(cls):
        from pint import _DEFAULT_REGISTRY

        cls.ureg = _DEFAULT_REGISTRY
        cls.U_ = cls.ureg.Unit
        cls.Q_ = cls.ureg.Quantity


class TestDimensionsWithDefaultRegistry(TestDimensions):
    @classmethod
    def setup_class(cls):
        from pint import _DEFAULT_REGISTRY

        cls.ureg = _DEFAULT_REGISTRY
        cls.Q_ = cls.ureg.Quantity


# TODO: do not subclass from QuantityTestCase
class TestOffsetUnitMath(QuantityTestCase):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.ureg.autoconvert_offset_to_baseunit = False
        cls.ureg.default_as_delta = True

    additions = [
        # --- input tuple -------------------- | -- expected result --
        (((100, "kelvin"), (10, "kelvin")), (110, "kelvin")),
        (((100, "kelvin"), (10, "degC")), "error"),
        (((100, "kelvin"), (10, "degF")), "error"),
        (((100, "kelvin"), (10, "degR")), (105.56, "kelvin")),
        (((100, "kelvin"), (10, "delta_degC")), (110, "kelvin")),
        (((100, "kelvin"), (10, "delta_degF")), (105.56, "kelvin")),
        (((100, "degC"), (10, "kelvin")), "error"),
        (((100, "degC"), (10, "degC")), "error"),
        (((100, "degC"), (10, "degF")), "error"),
        (((100, "degC"), (10, "degR")), "error"),
        (((100, "degC"), (10, "delta_degC")), (110, "degC")),
        (((100, "degC"), (10, "delta_degF")), (105.56, "degC")),
        (((100, "degF"), (10, "kelvin")), "error"),
        (((100, "degF"), (10, "degC")), "error"),
        (((100, "degF"), (10, "degF")), "error"),
        (((100, "degF"), (10, "degR")), "error"),
        (((100, "degF"), (10, "delta_degC")), (118, "degF")),
        (((100, "degF"), (10, "delta_degF")), (110, "degF")),
        (((100, "degR"), (10, "kelvin")), (118, "degR")),
        (((100, "degR"), (10, "degC")), "error"),
        (((100, "degR"), (10, "degF")), "error"),
        (((100, "degR"), (10, "degR")), (110, "degR")),
        (((100, "degR"), (10, "delta_degC")), (118, "degR")),
        (((100, "degR"), (10, "delta_degF")), (110, "degR")),
        (((100, "delta_degC"), (10, "kelvin")), (110, "kelvin")),
        (((100, "delta_degC"), (10, "degC")), (110, "degC")),
        (((100, "delta_degC"), (10, "degF")), (190, "degF")),
        (((100, "delta_degC"), (10, "degR")), (190, "degR")),
        (((100, "delta_degC"), (10, "delta_degC")), (110, "delta_degC")),
        (((100, "delta_degC"), (10, "delta_degF")), (105.56, "delta_degC")),
        (((100, "delta_degF"), (10, "kelvin")), (65.56, "kelvin")),
        (((100, "delta_degF"), (10, "degC")), (65.56, "degC")),
        (((100, "delta_degF"), (10, "degF")), (110, "degF")),
        (((100, "delta_degF"), (10, "degR")), (110, "degR")),
        (((100, "delta_degF"), (10, "delta_degC")), (118, "delta_degF")),
        (((100, "delta_degF"), (10, "delta_degF")), (110, "delta_degF")),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), additions)
    def test_addition(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        qin1, qin2 = input_tuple
        q1, q2 = self.Q_(*qin1), self.Q_(*qin2)
        # update input tuple with new values to have correct values on failure
        input_tuple = q1, q2
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.add(q1, q2)
        else:
            expected = self.Q_(*expected)
            assert op.add(q1, q2).units == expected.units
            helpers.assert_quantity_almost_equal(op.add(q1, q2), expected, atol=0.01)

    @helpers.requires_numpy
    @pytest.mark.parametrize(("input_tuple", "expected"), additions)
    def test_inplace_addition(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        (q1v, q1u), (q2v, q2u) = input_tuple
        # update input tuple with new values to have correct values on failure
        input_tuple = (
            (np.array([q1v] * 2, dtype=float), q1u),
            (np.array([q2v] * 2, dtype=float), q2u),
        )
        Q_ = self.Q_
        qin1, qin2 = input_tuple
        q1, q2 = Q_(*qin1), Q_(*qin2)
        q1_cp = copy.copy(q1)
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.iadd(q1_cp, q2)
        else:
            expected = np.array([expected[0]] * 2, dtype=float), expected[1]
            assert op.iadd(q1_cp, q2).units == Q_(*expected).units
            q1_cp = copy.copy(q1)
            helpers.assert_quantity_almost_equal(
                op.iadd(q1_cp, q2), Q_(*expected), atol=0.01
            )

    subtractions = [
        (((100, "kelvin"), (10, "kelvin")), (90, "kelvin")),
        (((100, "kelvin"), (10, "degC")), (-183.15, "kelvin")),
        (((100, "kelvin"), (10, "degF")), (-160.93, "kelvin")),
        (((100, "kelvin"), (10, "degR")), (94.44, "kelvin")),
        (((100, "kelvin"), (10, "delta_degC")), (90, "kelvin")),
        (((100, "kelvin"), (10, "delta_degF")), (94.44, "kelvin")),
        (((100, "degC"), (10, "kelvin")), (363.15, "delta_degC")),
        (((100, "degC"), (10, "degC")), (90, "delta_degC")),
        (((100, "degC"), (10, "degF")), (112.22, "delta_degC")),
        (((100, "degC"), (10, "degR")), (367.59, "delta_degC")),
        (((100, "degC"), (10, "delta_degC")), (90, "degC")),
        (((100, "degC"), (10, "delta_degF")), (94.44, "degC")),
        (((100, "degF"), (10, "kelvin")), (541.67, "delta_degF")),
        (((100, "degF"), (10, "degC")), (50, "delta_degF")),
        (((100, "degF"), (10, "degF")), (90, "delta_degF")),
        (((100, "degF"), (10, "degR")), (549.67, "delta_degF")),
        (((100, "degF"), (10, "delta_degC")), (82, "degF")),
        (((100, "degF"), (10, "delta_degF")), (90, "degF")),
        (((100, "degR"), (10, "kelvin")), (82, "degR")),
        (((100, "degR"), (10, "degC")), (-409.67, "degR")),
        (((100, "degR"), (10, "degF")), (-369.67, "degR")),
        (((100, "degR"), (10, "degR")), (90, "degR")),
        (((100, "degR"), (10, "delta_degC")), (82, "degR")),
        (((100, "degR"), (10, "delta_degF")), (90, "degR")),
        (((100, "delta_degC"), (10, "kelvin")), (90, "kelvin")),
        (((100, "delta_degC"), (10, "degC")), (90, "degC")),
        (((100, "delta_degC"), (10, "degF")), (170, "degF")),
        (((100, "delta_degC"), (10, "degR")), (170, "degR")),
        (((100, "delta_degC"), (10, "delta_degC")), (90, "delta_degC")),
        (((100, "delta_degC"), (10, "delta_degF")), (94.44, "delta_degC")),
        (((100, "delta_degF"), (10, "kelvin")), (45.56, "kelvin")),
        (((100, "delta_degF"), (10, "degC")), (45.56, "degC")),
        (((100, "delta_degF"), (10, "degF")), (90, "degF")),
        (((100, "delta_degF"), (10, "degR")), (90, "degR")),
        (((100, "delta_degF"), (10, "delta_degC")), (82, "delta_degF")),
        (((100, "delta_degF"), (10, "delta_degF")), (90, "delta_degF")),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), subtractions)
    def test_subtraction(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        qin1, qin2 = input_tuple
        q1, q2 = self.Q_(*qin1), self.Q_(*qin2)
        input_tuple = q1, q2
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.sub(q1, q2)
        else:
            expected = self.Q_(*expected)
            assert op.sub(q1, q2).units == expected.units
            helpers.assert_quantity_almost_equal(op.sub(q1, q2), expected, atol=0.01)

    #    @pytest.mark.xfail
    @helpers.requires_numpy
    @pytest.mark.parametrize(("input_tuple", "expected"), subtractions)
    def test_inplace_subtraction(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        (q1v, q1u), (q2v, q2u) = input_tuple
        # update input tuple with new values to have correct values on failure
        input_tuple = (
            (np.array([q1v] * 2, dtype=float), q1u),
            (np.array([q2v] * 2, dtype=float), q2u),
        )
        Q_ = self.Q_
        qin1, qin2 = input_tuple
        q1, q2 = Q_(*qin1), Q_(*qin2)
        q1_cp = copy.copy(q1)
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.isub(q1_cp, q2)
        else:
            expected = np.array([expected[0]] * 2, dtype=float), expected[1]
            assert op.isub(q1_cp, q2).units == Q_(*expected).units
            q1_cp = copy.copy(q1)
            helpers.assert_quantity_almost_equal(
                op.isub(q1_cp, q2), Q_(*expected), atol=0.01
            )

    multiplications = [
        (((100, "kelvin"), (10, "kelvin")), (1000, "kelvin**2")),
        (((100, "kelvin"), (10, "degC")), "error"),
        (((100, "kelvin"), (10, "degF")), "error"),
        (((100, "kelvin"), (10, "degR")), (1000, "kelvin*degR")),
        (((100, "kelvin"), (10, "delta_degC")), (1000, "kelvin*delta_degC")),
        (((100, "kelvin"), (10, "delta_degF")), (1000, "kelvin*delta_degF")),
        (((100, "degC"), (10, "kelvin")), "error"),
        (((100, "degC"), (10, "degC")), "error"),
        (((100, "degC"), (10, "degF")), "error"),
        (((100, "degC"), (10, "degR")), "error"),
        (((100, "degC"), (10, "delta_degC")), "error"),
        (((100, "degC"), (10, "delta_degF")), "error"),
        (((100, "degF"), (10, "kelvin")), "error"),
        (((100, "degF"), (10, "degC")), "error"),
        (((100, "degF"), (10, "degF")), "error"),
        (((100, "degF"), (10, "degR")), "error"),
        (((100, "degF"), (10, "delta_degC")), "error"),
        (((100, "degF"), (10, "delta_degF")), "error"),
        (((100, "degR"), (10, "kelvin")), (1000, "degR*kelvin")),
        (((100, "degR"), (10, "degC")), "error"),
        (((100, "degR"), (10, "degF")), "error"),
        (((100, "degR"), (10, "degR")), (1000, "degR**2")),
        (((100, "degR"), (10, "delta_degC")), (1000, "degR*delta_degC")),
        (((100, "degR"), (10, "delta_degF")), (1000, "degR*delta_degF")),
        (((100, "delta_degC"), (10, "kelvin")), (1000, "delta_degC*kelvin")),
        (((100, "delta_degC"), (10, "degC")), "error"),
        (((100, "delta_degC"), (10, "degF")), "error"),
        (((100, "delta_degC"), (10, "degR")), (1000, "delta_degC*degR")),
        (((100, "delta_degC"), (10, "delta_degC")), (1000, "delta_degC**2")),
        (((100, "delta_degC"), (10, "delta_degF")), (1000, "delta_degC*delta_degF")),
        (((100, "delta_degF"), (10, "kelvin")), (1000, "delta_degF*kelvin")),
        (((100, "delta_degF"), (10, "degC")), "error"),
        (((100, "delta_degF"), (10, "degF")), "error"),
        (((100, "delta_degF"), (10, "degR")), (1000, "delta_degF*degR")),
        (((100, "delta_degF"), (10, "delta_degC")), (1000, "delta_degF*delta_degC")),
        (((100, "delta_degF"), (10, "delta_degF")), (1000, "delta_degF**2")),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), multiplications)
    def test_multiplication(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        qin1, qin2 = input_tuple
        q1, q2 = self.Q_(*qin1), self.Q_(*qin2)
        input_tuple = q1, q2
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.mul(q1, q2)
        else:
            expected = self.Q_(*expected)
            assert op.mul(q1, q2).units == expected.units
            helpers.assert_quantity_almost_equal(op.mul(q1, q2), expected, atol=0.01)

    @helpers.requires_numpy
    @pytest.mark.parametrize(("input_tuple", "expected"), multiplications)
    def test_inplace_multiplication(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        (q1v, q1u), (q2v, q2u) = input_tuple
        # update input tuple with new values to have correct values on failure
        input_tuple = (
            (np.array([q1v] * 2, dtype=float), q1u),
            (np.array([q2v] * 2, dtype=float), q2u),
        )
        Q_ = self.Q_
        qin1, qin2 = input_tuple
        q1, q2 = Q_(*qin1), Q_(*qin2)
        q1_cp = copy.copy(q1)
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.imul(q1_cp, q2)
        else:
            expected = np.array([expected[0]] * 2, dtype=float), expected[1]
            assert op.imul(q1_cp, q2).units == Q_(*expected).units
            q1_cp = copy.copy(q1)
            helpers.assert_quantity_almost_equal(
                op.imul(q1_cp, q2), Q_(*expected), atol=0.01
            )

    divisions = [
        (((100, "kelvin"), (10, "kelvin")), (10, "")),
        (((100, "kelvin"), (10, "degC")), "error"),
        (((100, "kelvin"), (10, "degF")), "error"),
        (((100, "kelvin"), (10, "degR")), (10, "kelvin/degR")),
        (((100, "kelvin"), (10, "delta_degC")), (10, "kelvin/delta_degC")),
        (((100, "kelvin"), (10, "delta_degF")), (10, "kelvin/delta_degF")),
        (((100, "degC"), (10, "kelvin")), "error"),
        (((100, "degC"), (10, "degC")), "error"),
        (((100, "degC"), (10, "degF")), "error"),
        (((100, "degC"), (10, "degR")), "error"),
        (((100, "degC"), (10, "delta_degC")), "error"),
        (((100, "degC"), (10, "delta_degF")), "error"),
        (((100, "degF"), (10, "kelvin")), "error"),
        (((100, "degF"), (10, "degC")), "error"),
        (((100, "degF"), (10, "degF")), "error"),
        (((100, "degF"), (10, "degR")), "error"),
        (((100, "degF"), (10, "delta_degC")), "error"),
        (((100, "degF"), (10, "delta_degF")), "error"),
        (((100, "degR"), (10, "kelvin")), (10, "degR/kelvin")),
        (((100, "degR"), (10, "degC")), "error"),
        (((100, "degR"), (10, "degF")), "error"),
        (((100, "degR"), (10, "degR")), (10, "")),
        (((100, "degR"), (10, "delta_degC")), (10, "degR/delta_degC")),
        (((100, "degR"), (10, "delta_degF")), (10, "degR/delta_degF")),
        (((100, "delta_degC"), (10, "kelvin")), (10, "delta_degC/kelvin")),
        (((100, "delta_degC"), (10, "degC")), "error"),
        (((100, "delta_degC"), (10, "degF")), "error"),
        (((100, "delta_degC"), (10, "degR")), (10, "delta_degC/degR")),
        (((100, "delta_degC"), (10, "delta_degC")), (10, "")),
        (((100, "delta_degC"), (10, "delta_degF")), (10, "delta_degC/delta_degF")),
        (((100, "delta_degF"), (10, "kelvin")), (10, "delta_degF/kelvin")),
        (((100, "delta_degF"), (10, "degC")), "error"),
        (((100, "delta_degF"), (10, "degF")), "error"),
        (((100, "delta_degF"), (10, "degR")), (10, "delta_degF/degR")),
        (((100, "delta_degF"), (10, "delta_degC")), (10, "delta_degF/delta_degC")),
        (((100, "delta_degF"), (10, "delta_degF")), (10, "")),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), divisions)
    def test_truedivision(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        qin1, qin2 = input_tuple
        q1, q2 = self.Q_(*qin1), self.Q_(*qin2)
        input_tuple = q1, q2
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.truediv(q1, q2)
        else:
            expected = self.Q_(*expected)
            assert op.truediv(q1, q2).units == expected.units
            helpers.assert_quantity_almost_equal(
                op.truediv(q1, q2), expected, atol=0.01
            )

    @helpers.requires_numpy
    @pytest.mark.parametrize(("input_tuple", "expected"), divisions)
    def test_inplace_truedivision(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = False
        (q1v, q1u), (q2v, q2u) = input_tuple
        # update input tuple with new values to have correct values on failure
        input_tuple = (
            (np.array([q1v] * 2, dtype=float), q1u),
            (np.array([q2v] * 2, dtype=float), q2u),
        )
        Q_ = self.Q_
        qin1, qin2 = input_tuple
        q1, q2 = Q_(*qin1), Q_(*qin2)
        q1_cp = copy.copy(q1)
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.itruediv(q1_cp, q2)
        else:
            expected = np.array([expected[0]] * 2, dtype=float), expected[1]
            assert op.itruediv(q1_cp, q2).units == Q_(*expected).units
            q1_cp = copy.copy(q1)
            helpers.assert_quantity_almost_equal(
                op.itruediv(q1_cp, q2), Q_(*expected), atol=0.01
            )

    multiplications_with_autoconvert_to_baseunit = [
        (((100, "kelvin"), (10, "degC")), (28315.0, "kelvin**2")),
        (((100, "kelvin"), (10, "degF")), (26092.78, "kelvin**2")),
        (((100, "degC"), (10, "kelvin")), (3731.5, "kelvin**2")),
        (((100, "degC"), (10, "degC")), (105657.42, "kelvin**2")),
        (((100, "degC"), (10, "degF")), (97365.20, "kelvin**2")),
        (((100, "degC"), (10, "degR")), (3731.5, "kelvin*degR")),
        (((100, "degC"), (10, "delta_degC")), (3731.5, "kelvin*delta_degC")),
        (((100, "degC"), (10, "delta_degF")), (3731.5, "kelvin*delta_degF")),
        (((100, "degF"), (10, "kelvin")), (3109.28, "kelvin**2")),
        (((100, "degF"), (10, "degC")), (88039.20, "kelvin**2")),
        (((100, "degF"), (10, "degF")), (81129.69, "kelvin**2")),
        (((100, "degF"), (10, "degR")), (3109.28, "kelvin*degR")),
        (((100, "degF"), (10, "delta_degC")), (3109.28, "kelvin*delta_degC")),
        (((100, "degF"), (10, "delta_degF")), (3109.28, "kelvin*delta_degF")),
        (((100, "degR"), (10, "degC")), (28315.0, "degR*kelvin")),
        (((100, "degR"), (10, "degF")), (26092.78, "degR*kelvin")),
        (((100, "delta_degC"), (10, "degC")), (28315.0, "delta_degC*kelvin")),
        (((100, "delta_degC"), (10, "degF")), (26092.78, "delta_degC*kelvin")),
        (((100, "delta_degF"), (10, "degC")), (28315.0, "delta_degF*kelvin")),
        (((100, "delta_degF"), (10, "degF")), (26092.78, "delta_degF*kelvin")),
    ]

    @pytest.mark.parametrize(
        ("input_tuple", "expected"), multiplications_with_autoconvert_to_baseunit
    )
    def test_multiplication_with_autoconvert(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = True
        qin1, qin2 = input_tuple
        q1, q2 = self.Q_(*qin1), self.Q_(*qin2)
        input_tuple = q1, q2
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.mul(q1, q2)
        else:
            expected = self.Q_(*expected)
            assert op.mul(q1, q2).units == expected.units
            helpers.assert_quantity_almost_equal(op.mul(q1, q2), expected, atol=0.01)

    @helpers.requires_numpy
    @pytest.mark.parametrize(
        ("input_tuple", "expected"), multiplications_with_autoconvert_to_baseunit
    )
    def test_inplace_multiplication_with_autoconvert(self, input_tuple, expected):
        self.ureg.autoconvert_offset_to_baseunit = True
        (q1v, q1u), (q2v, q2u) = input_tuple
        # update input tuple with new values to have correct values on failure
        input_tuple = (
            (np.array([q1v] * 2, dtype=float), q1u),
            (np.array([q2v] * 2, dtype=float), q2u),
        )
        Q_ = self.Q_
        qin1, qin2 = input_tuple
        q1, q2 = Q_(*qin1), Q_(*qin2)
        q1_cp = copy.copy(q1)
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.imul(q1_cp, q2)
        else:
            expected = np.array([expected[0]] * 2, dtype=float), expected[1]
            assert op.imul(q1_cp, q2).units == Q_(*expected).units
            q1_cp = copy.copy(q1)
            helpers.assert_quantity_almost_equal(
                op.imul(q1_cp, q2), Q_(*expected), atol=0.01
            )

    multiplications_with_scalar = [
        (((10, "kelvin"), 2), (20.0, "kelvin")),
        (((10, "kelvin**2"), 2), (20.0, "kelvin**2")),
        (((10, "degC"), 2), (20.0, "degC")),
        (((10, "1/degC"), 2), "error"),
        (((10, "degC**0.5"), 2), "error"),
        (((10, "degC**2"), 2), "error"),
        (((10, "degC**-2"), 2), "error"),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), multiplications_with_scalar)
    def test_multiplication_with_scalar(self, input_tuple, expected):
        self.ureg.default_as_delta = False
        in1, in2 = input_tuple
        if type(in1) is tuple:
            in1, in2 = self.Q_(*in1), in2
        else:
            in1, in2 = in1, self.Q_(*in2)
        input_tuple = in1, in2  # update input_tuple for better tracebacks
        if expected == "error":
            with pytest.raises(OffsetUnitCalculusError):
                op.mul(in1, in2)
        else:
            expected = self.Q_(*expected)
            assert op.mul(in1, in2).units == expected.units
            helpers.assert_quantity_almost_equal(op.mul(in1, in2), expected, atol=0.01)

    divisions_with_scalar = [  # without / with autoconvert to plain unit
        (((10, "kelvin"), 2), [(5.0, "kelvin"), (5.0, "kelvin")]),
        (((10, "kelvin**2"), 2), [(5.0, "kelvin**2"), (5.0, "kelvin**2")]),
        (((10, "degC"), 2), ["error", "error"]),
        (((10, "degC**2"), 2), ["error", "error"]),
        (((10, "degC**-2"), 2), ["error", "error"]),
        ((2, (10, "kelvin")), [(0.2, "1/kelvin"), (0.2, "1/kelvin")]),
        ((2, (10, "degC")), ["error", (2 / 283.15, "1/kelvin")]),
        ((2, (10, "degC**2")), ["error", "error"]),
        ((2, (10, "degC**-2")), ["error", "error"]),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), divisions_with_scalar)
    def test_division_with_scalar(self, input_tuple, expected):
        self.ureg.default_as_delta = False
        in1, in2 = input_tuple
        if type(in1) is tuple:
            in1, in2 = self.Q_(*in1), in2
        else:
            in1, in2 = in1, self.Q_(*in2)
        input_tuple = in1, in2  # update input_tuple for better tracebacks
        expected_copy = expected.copy()
        for i, mode in enumerate([False, True]):
            self.ureg.autoconvert_offset_to_baseunit = mode
            if expected_copy[i] == "error":
                with pytest.raises(OffsetUnitCalculusError):
                    op.truediv(in1, in2)
            else:
                expected = self.Q_(*expected_copy[i])
                assert op.truediv(in1, in2).units == expected.units
                helpers.assert_quantity_almost_equal(op.truediv(in1, in2), expected)

    exponentiation = [  # results without / with autoconvert
        (((10, "degC"), 1), [(10, "degC"), (10, "degC")]),
        (((10, "degC"), 0.5), ["error", (283.15**0.5, "kelvin**0.5")]),
        (((10, "degC"), 0), [(1.0, ""), (1.0, "")]),
        (((10, "degC"), -1), ["error", (1 / (10 + 273.15), "kelvin**-1")]),
        (((10, "degC"), -2), ["error", (1 / (10 + 273.15) ** 2.0, "kelvin**-2")]),
        (((0, "degC"), -2), ["error", (1 / 273.15**2, "kelvin**-2")]),
        (((10, "degC"), (2, "")), ["error", (283.15**2, "kelvin**2")]),
        (((10, "degC"), (10, "degK")), ["error", "error"]),
        (((10, "kelvin"), (2, "")), [(100.0, "kelvin**2"), (100.0, "kelvin**2")]),
        ((2, (2, "kelvin")), ["error", "error"]),
        ((2, (500.0, "millikelvin/kelvin")), [2**0.5, 2**0.5]),
        ((2, (0.5, "kelvin/kelvin")), [2**0.5, 2**0.5]),
        (
            ((10, "degC"), (500.0, "millikelvin/kelvin")),
            ["error", (283.15**0.5, "kelvin**0.5")],
        ),
    ]

    @pytest.mark.parametrize(("input_tuple", "expected"), exponentiation)
    def test_exponentiation(self, input_tuple, expected):
        self.ureg.default_as_delta = False
        in1, in2 = input_tuple
        if type(in1) is type(in2) is tuple:
            in1, in2 = self.Q_(*in1), self.Q_(*in2)
        elif type(in1) is not tuple and type(in2) is tuple:
            in2 = self.Q_(*in2)
        else:
            in1 = self.Q_(*in1)
        input_tuple = in1, in2
        expected_copy = expected.copy()
        for i, mode in enumerate([False, True]):
            self.ureg.autoconvert_offset_to_baseunit = mode
            if expected_copy[i] == "error":
                with pytest.raises((OffsetUnitCalculusError, DimensionalityError)):
                    op.pow(in1, in2)
            else:
                if type(expected_copy[i]) is tuple:
                    expected = self.Q_(*expected_copy[i])
                    assert op.pow(in1, in2).units == expected.units
                else:
                    expected = expected_copy[i]
                helpers.assert_quantity_almost_equal(op.pow(in1, in2), expected)

    @helpers.requires_numpy
    def test_exponentiation_force_ndarray(self):
        ureg = UnitRegistry(force_ndarray_like=True)
        q = ureg.Quantity(1, "1 / hours")

        q1 = q**2
        assert all(isinstance(v, int) for v in q1._units.values())

        q2 = q.copy()
        q2 **= 2
        assert all(isinstance(v, int) for v in q2._units.values())

    @helpers.requires_numpy
    @pytest.mark.parametrize(("input_tuple", "expected"), exponentiation)
    def test_inplace_exponentiation(self, input_tuple, expected):
        self.ureg.default_as_delta = False
        in1, in2 = input_tuple
        if type(in1) is type(in2) is tuple:
            (q1v, q1u), (q2v, q2u) = in1, in2
            in1 = self.Q_(*(np.array([q1v] * 2, dtype=float), q1u))
            in2 = self.Q_(q2v, q2u)
        elif type(in1) is not tuple and type(in2) is tuple:
            in2 = self.Q_(*in2)
        else:
            in1 = self.Q_(*in1)

        input_tuple = in1, in2

        expected_copy = expected.copy()
        for i, mode in enumerate([False, True]):
            self.ureg.autoconvert_offset_to_baseunit = mode
            in1_cp = copy.copy(in1)
            if expected_copy[i] == "error":
                with pytest.raises((OffsetUnitCalculusError, DimensionalityError)):
                    op.ipow(in1_cp, in2)
            else:
                if type(expected_copy[i]) is tuple:
                    expected = self.Q_(
                        np.array([expected_copy[i][0]] * 2, dtype=float),
                        expected_copy[i][1],
                    )
                    assert op.ipow(in1_cp, in2).units == expected.units
                else:
                    expected = np.array([expected_copy[i]] * 2, dtype=float)

                in1_cp = copy.copy(in1)
                helpers.assert_quantity_almost_equal(op.ipow(in1_cp, in2), expected)

    # matmul is only a ufunc since 1.16
    @helpers.requires_numpy_at_least("1.16")
    def test_matmul_with_numpy(self):
        A = [[1, 2], [3, 4]] * self.ureg.m
        B = np.array([[0, -1], [-1, 0]])
        b = [[1], [0]] * self.ureg.m
        helpers.assert_quantity_equal(A @ B, [[-2, -1], [-4, -3]] * self.ureg.m)
        helpers.assert_quantity_equal(A @ b, [[1], [3]] * self.ureg.m**2)
        helpers.assert_quantity_equal(B @ b, [[0], [-1]] * self.ureg.m)


class TestDimensionReduction:
    def _calc_mass(self, ureg):
        density = 3 * ureg.g / ureg.L
        volume = 32 * ureg.milliliter
        return density * volume

    def _icalc_mass(self, ureg):
        res = ureg.Quantity(3.0, "gram/liter")
        res *= ureg.Quantity(32.0, "milliliter")
        return res

    def test_mul_and_div_reduction(self):
        ureg = UnitRegistry(auto_reduce_dimensions=True)
        mass = self._calc_mass(ureg)
        assert mass.units == ureg.g
        ureg = UnitRegistry(auto_reduce_dimensions=False)
        mass = self._calc_mass(ureg)
        assert mass.units == ureg.g / ureg.L * ureg.milliliter

    @helpers.requires_numpy
    def test_imul_and_div_reduction(self):
        ureg = UnitRegistry(auto_reduce_dimensions=True, force_ndarray=True)
        mass = self._icalc_mass(ureg)
        assert mass.units == ureg.g
        ureg = UnitRegistry(auto_reduce_dimensions=False, force_ndarray=True)
        mass = self._icalc_mass(ureg)
        assert mass.units == ureg.g / ureg.L * ureg.milliliter

    def test_reduction_to_dimensionless(self):
        ureg = UnitRegistry(auto_reduce_dimensions=True)
        x = (10 * ureg.feet) / (3 * ureg.inches)
        assert x.units == UnitsContainer({})
        ureg = UnitRegistry(auto_reduce_dimensions=False)
        x = (10 * ureg.feet) / (3 * ureg.inches)
        assert x.units == ureg.feet / ureg.inches

    def test_nocoerce_creation(self):
        ureg = UnitRegistry(auto_reduce_dimensions=True)
        x = 1 * ureg.foot
        assert x.units == ureg.foot


# TODO: do not subclass from QuantityTestCase
class TestTimedelta(QuantityTestCase):
    def test_add_sub(self):
        d = datetime.datetime(year=1968, month=1, day=10, hour=3, minute=42, second=24)
        after = d + 3 * self.ureg.second
        assert d + datetime.timedelta(seconds=3) == after
        after = 3 * self.ureg.second + d
        assert d + datetime.timedelta(seconds=3) == after
        after = d - 3 * self.ureg.second
        assert d - datetime.timedelta(seconds=3) == after
        with pytest.raises(DimensionalityError):
            3 * self.ureg.second - d

    def test_iadd_isub(self):
        d = datetime.datetime(year=1968, month=1, day=10, hour=3, minute=42, second=24)
        after = copy.copy(d)
        after += 3 * self.ureg.second
        assert d + datetime.timedelta(seconds=3) == after
        after = 3 * self.ureg.second
        after += d
        assert d + datetime.timedelta(seconds=3) == after
        after = copy.copy(d)
        after -= 3 * self.ureg.second
        assert d - datetime.timedelta(seconds=3) == after
        after = 3 * self.ureg.second
        with pytest.raises(DimensionalityError):
            after -= d


# TODO: do not subclass from QuantityTestCase
class TestCompareNeutral(QuantityTestCase):
    """Test comparisons against non-Quantity zero or NaN values for for
    non-dimensionless quantities
    """

    def test_equal_zero(self):
        self.ureg.autoconvert_offset_to_baseunit = False
        assert self.Q_(0, "J") == 0
        assert not (self.Q_(0, "J") == self.Q_(0, ""))
        assert not (self.Q_(5, "J") == 0)

    def test_equal_nan(self):
        # nan == nan returns False
        self.ureg.autoconvert_offset_to_baseunit = False
        assert not (self.Q_(math.nan, "J") == 0)
        assert not (self.Q_(math.nan, "J") == math.nan)
        assert not (self.Q_(math.nan, "J") == self.Q_(math.nan, ""))
        assert not (self.Q_(5, "J") == math.nan)

    @helpers.requires_numpy
    def test_equal_zero_nan_NP(self):
        self.ureg.autoconvert_offset_to_baseunit = False
        aeq = np.testing.assert_array_equal
        aeq(self.Q_(0, "J") == np.array([0, np.nan]), np.array([True, False]))
        aeq(self.Q_(5, "J") == np.array([0, np.nan]), np.array([False, False]))
        aeq(
            self.Q_([0, 1, 2], "J") == np.array([0, 0, np.nan]),
            np.asarray([True, False, False]),
        )

        # This raise an exception on NumPy 1.25 as dimensions
        # are different
        # assert not (self.Q_(np.arange(4), "J") == np.zeros(3))

    def test_offset_equal_zero(self):
        ureg = self.ureg
        ureg.autoconvert_offset_to_baseunit = False
        q0 = ureg.Quantity(-273.15, "degC")
        q1 = ureg.Quantity(0, "degC")
        q2 = ureg.Quantity(5, "degC")
        with pytest.raises(OffsetUnitCalculusError):
            q0.__eq__(0)
        with pytest.raises(OffsetUnitCalculusError):
            q1.__eq__(0)
        with pytest.raises(OffsetUnitCalculusError):
            q2.__eq__(0)
        assert not (q0 == ureg.Quantity(0, ""))

    def test_offset_autoconvert_equal_zero(self):
        ureg = self.ureg
        ureg.autoconvert_offset_to_baseunit = True
        q0 = ureg.Quantity(-273.15, "degC")
        q1 = ureg.Quantity(0, "degC")
        q2 = ureg.Quantity(5, "degC")
        assert q0 == 0
        assert not (q1 == 0)
        assert not (q2 == 0)
        assert not (q0 == ureg.Quantity(0, ""))

    def test_gt_zero(self):
        self.ureg.autoconvert_offset_to_baseunit = False
        q0 = self.Q_(0, "J")
        q0m = self.Q_(0, "m")
        q0less = self.Q_(0, "")
        qpos = self.Q_(5, "J")
        qneg = self.Q_(-5, "J")
        assert qpos > q0
        assert qpos > 0
        assert not (qneg > 0)
        with pytest.raises(DimensionalityError):
            qpos > q0less
        with pytest.raises(DimensionalityError):
            qpos > q0m

    def test_gt_nan(self):
        self.ureg.autoconvert_offset_to_baseunit = False
        qn = self.Q_(math.nan, "J")
        qnm = self.Q_(math.nan, "m")
        qnless = self.Q_(math.nan, "")
        qpos = self.Q_(5, "J")
        assert not (qpos > qn)
        assert not (qpos > math.nan)
        with pytest.raises(DimensionalityError):
            qpos > qnless
        with pytest.raises(DimensionalityError):
            qpos > qnm

    @helpers.requires_numpy
    def test_gt_zero_nan_NP(self):
        self.ureg.autoconvert_offset_to_baseunit = False
        qpos = self.Q_(5, "J")
        qneg = self.Q_(-5, "J")
        aeq = np.testing.assert_array_equal
        aeq(qpos > np.array([0, np.nan]), np.asarray([True, False]))
        aeq(qneg > np.array([0, np.nan]), np.asarray([False, False]))
        aeq(
            self.Q_(np.arange(-2, 3), "J") > np.array([np.nan, 0, 0, 0, np.nan]),
            np.asarray([False, False, False, True, False]),
        )
        with pytest.raises(ValueError):
            self.Q_(np.arange(-1, 2), "J") > np.zeros(4)

    def test_offset_gt_zero(self):
        ureg = self.ureg
        ureg.autoconvert_offset_to_baseunit = False
        q0 = ureg.Quantity(-273.15, "degC")
        q1 = ureg.Quantity(0, "degC")
        q2 = ureg.Quantity(5, "degC")
        with pytest.raises(OffsetUnitCalculusError):
            q0.__gt__(0)
        with pytest.raises(OffsetUnitCalculusError):
            q1.__gt__(0)
        with pytest.raises(OffsetUnitCalculusError):
            q2.__gt__(0)
        with pytest.raises(DimensionalityError):
            q1.__gt__(ureg.Quantity(0, ""))

    def test_offset_autoconvert_gt_zero(self):
        ureg = self.ureg
        ureg.autoconvert_offset_to_baseunit = True
        q0 = ureg.Quantity(-273.15, "degC")
        q1 = ureg.Quantity(0, "degC")
        q2 = ureg.Quantity(5, "degC")
        assert not (q0 > 0)
        assert q1 > 0
        assert q2 > 0
        with pytest.raises(DimensionalityError):
            q1.__gt__(ureg.Quantity(0, ""))

    def test_types(self):
        quantity = self.Q_(1.0, "m")
        assert isinstance(quantity, self.Q_)
        assert isinstance(quantity.units, self.ureg.Unit)
        assert isinstance(quantity.m, float)

        assert isinstance(self.ureg.m, self.ureg.Unit)
